import java.io.*;
import java.util.Vector;



class Atom{
	double[] R;
	String atom_name;
	int seq_num;
	String pdb_line;
	
	public Atom(){		
		R = new double[3];	
	}
}




public class AssignOGAtoms{
	public static void main( String[] args ) throws IOException{
		if( args.length == 0 || args[0].equalsIgnoreCase( "--help" ) ){
				System.out.println( "Usage:   AssignOGAtoms" );
				System.out.println( "              -opts  [PDB glycan structure file]" );
				
				return;
		}	
		
		FileInputStream list_stream;
		BufferedReader list_reader;	
		String list_line;
		
		FileInputStream struct_stream;
		BufferedReader struct_reader;	
		String struct_line;

		DataOutputStream out;		
		
		String mStruct = args[0];
		
		
		
		String [] array;		
		Atom atom;
		Atom atom_i, atom_j;
		double distance;
		String atom_name = "";
		int num_atoms;
		
		int num_wrong_structures = 0;
		

		

		try{
			Vector<Atom> vAtoms = new Vector<Atom>();
			
			struct_stream = new FileInputStream( new File( mStruct ) );
			struct_reader = new BufferedReader( new InputStreamReader( struct_stream ) );		
			
			while( ( struct_line = struct_reader.readLine()) != null ){
				if( struct_line.startsWith( "ATOM" ) || struct_line.startsWith( "HETATM" ) ){
					atom = new Atom();
					atom.atom_name = struct_line.substring( 12, 16 ).trim();
					atom.seq_num = Integer.parseInt( struct_line.substring( 22, 26 ).trim() );
					atom.R[0] = Double.parseDouble( struct_line.substring( 30, 38 ).trim() );
					atom.R[1] = Double.parseDouble( struct_line.substring( 38, 46 ).trim() );
					atom.R[2] = Double.parseDouble( struct_line.substring( 46, 54 ).trim() );
					
					atom.pdb_line = struct_line;
					
					vAtoms.add( atom );
				}
			}
			
			struct_reader.close();
			struct_stream.close();
			
			num_atoms = vAtoms.size();
			
			
			
			
			
			// =====================================
			//    Renumber glycan residue numbers
			// =====================================
			
			int prev_resSeq = -10000;
			int curr_resSeq;
			int renum_resSeq = 0;
			
			for( int i = 0 ; i < num_atoms ; i++ ){
				curr_resSeq = vAtoms.elementAt(i).seq_num;
				
				if( curr_resSeq != prev_resSeq ){
					renum_resSeq++;
				} 
				
				vAtoms.elementAt(i).seq_num = renum_resSeq;
				prev_resSeq = curr_resSeq;
			}
			
			String pdb_line_old = "";
			
			for( int i = 0 ; i < num_atoms ; i++ ){
				pdb_line_old = vAtoms.elementAt(i).pdb_line;
				
				if( vAtoms.elementAt(i).seq_num >= 1 && vAtoms.elementAt(i).seq_num < 10 )
					vAtoms.elementAt(i).pdb_line = pdb_line_old.substring( 0, 22 ) + "   " + vAtoms.elementAt(i).seq_num + pdb_line_old.substring( 26, pdb_line_old.length() );
				else if( vAtoms.elementAt(i).seq_num >= 10 && vAtoms.elementAt(i).seq_num < 100 )
					vAtoms.elementAt(i).pdb_line = pdb_line_old.substring( 0, 22 ) + "  " + vAtoms.elementAt(i).seq_num + pdb_line_old.substring( 26, pdb_line_old.length() );
				else
					vAtoms.elementAt(i).pdb_line = pdb_line_old.substring( 0, 22 ) + " " + vAtoms.elementAt(i).seq_num + pdb_line_old.substring( 26, pdb_line_old.length() );
			}
			

			
			
			// ===============================
			//    Assign glycosidic oxygens
			// ===============================
			
			int num_OG = 0;
			
			for( int i = 0 ; i < num_atoms ; i++ ){
				atom_i = vAtoms.elementAt(i);
				
				if( atom_i.atom_name.charAt(0) == 'O' ){
					for( int j = 0 ; j < num_atoms ; j++ ){
						atom_j = vAtoms.elementAt(j);
						
						if( ( atom_j.atom_name.length() > 1 ) && 
						   	( ( atom_j.atom_name.charAt(0) == 'C' && atom_j.atom_name.charAt(1) == '1' ) || ( atom_j.atom_name.charAt(0) == 'C' && atom_j.atom_name.charAt(1) == '2' ) ) &&
						   	( atom_i.seq_num != atom_j.seq_num ) ){
							distance = getDistance( atom_i.R, atom_j.R );
							
							if( distance < 2.0 ){
								num_OG++;
									
								if( atom_j.seq_num < 10 )
									atom_name = "OG" + atom_j.seq_num + " ";
							   	else
							    	atom_name = "OG" + atom_j.seq_num;
							    	
							   	atom_i.pdb_line = atom_i.pdb_line.substring( 0, 13 ) + atom_name + atom_i.pdb_line.substring( 17, atom_i.pdb_line.length() );
							    	
							   	break;
							}
						}
					}
				}
			}	
			
			
                
                             
            // ===================
			//    Write output
			// =================== 		
			
			System.out.println( "Number of residues: " + renum_resSeq );
			System.out.println( "Number of glycosidic oxygens: " + num_OG );
			
			array = mStruct.split( ".pdb" );
			
			out = new DataOutputStream( new FileOutputStream( new File( array[0] + "_OG.pdb" ) ) );	
			
			for( int i = 0 ; i < num_atoms ; i++ ){
				out.writeBytes( vAtoms.elementAt(i).pdb_line + "\n" );
			}
			
			out.writeBytes( "TER" );
			out.close();
			
			if( ( renum_resSeq > 1 ) && ( ( renum_resSeq - 1 ) != num_OG ) ){
				System.out.println( "Error: wrong number of glycosidic oxygens. please check original PDB file." );
			}
			
			vAtoms.removeAllElements();
		}
		catch( Exception exception ){
			exception.printStackTrace();
		}		
	}	
	
	
	
	static public double getDistance( double[] r1, double[] r2 ){
		return Math.sqrt( ( r1[0] - r2[0] ) * ( r1[0] - r2[0] ) + ( r1[1] - r2[1] ) * ( r1[1] - r2[1] ) + ( r1[2] - r2[2] ) * ( r1[2] - r2[2] ) );
	}
}
	
	
