#include <fstream>
#include <iostream>
#include <list>
#include <string>
#include <set>
#include <vector>
#include <queue>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cmath>
#include <map>
#include <string.h>
#include <algorithm>
#include <assert.h>

using namespace std;




// **************************
// **** Global Variables ****
// **************************

string progname;
string s1, s2, ref_struct, fit_struct;
int N_target_option = 1;
int output_option = 1;
float dist_tolerance = 1.0;
float d0_OG = 3.0;	
float d0_Ring = 3.0;

double u[3][3], t[3];
double u_best[3][3], t_best[3];

float shortestDist[20];
double cost[20][20];
double score[20][20];

int num_linkages[20];
int linked_residues[20][2];

int spalte[20];
int alignedRes[20];
int alignedpair[20];

float Ring_rmsd[100];
float OG_rmsd[100];







// *****************
// **** Classes ****
// *****************

class Atom{
	public:
		Atom();
	    ~Atom();
		Atom( Atom & );
		
	public:		
		char atom_name;  
		int resSeq;  
		double R[3];   
};

Atom :: Atom()
{}

Atom :: ~Atom()
{}

Atom :: Atom( Atom & )
{}



class Pair{
	public:
		Pair();
	    ~Pair();
		Pair( Pair & );
		
	public:		
		int ref_index;
	    int fit_index;
};

Pair :: Pair()
{}

Pair :: ~Pair()
{}

Pair :: Pair( Pair & )
{}



class Edge{
	public:
		Edge();
	    ~Edge();
		Edge( Edge & );
		
	public:		
		int pair_i;
	    int pair_j;
};

Edge :: Edge()
{}

Edge :: ~Edge()
{}

Edge :: Edge( Edge & )
{}



class Residue{
	public:
		Residue();
	    ~Residue();
		Residue( Residue & );
		
	public:		
		int resSeq;
		char resName[4];
		int resSeq_OG_linked;
		int hasOG;           // 1: This residue has OG atom
		double Ring[6][3];   // 0-C1, 1-C2, 2-C3, 3-C4, 4-C5, 5-O5
		double R[3];  
		double OG[3];
		
		double Ring_sup[6][3];
		double R_sup[3];  
		double OG_sup[3];
};

Residue :: Residue(){
	resSeq_OG_linked = 0; 
}

Residue :: ~Residue()
{}

Residue :: Residue( Residue & )
{}



class Fragment{
	public:
		Fragment();
	    ~Fragment();
		Fragment( Fragment & );
		
	public:		
		int index[4];		
};

Fragment :: Fragment()
{}

Fragment :: ~Fragment()
{}

Fragment :: Fragment( Fragment & )
{}







// *********************************
// ********* Parse Command *********
// *********************************

void help( string msg ){
  cout << msg
  << progname << " [options]                                             " << "\n"
  << "options:                                                           " << "\n"
  << "    -s1   glycan structure 1 (pdb)                                 " << "\n"
  << "    -s2   glycan structure 2 (pdb)                                 " << "\n"
  << "    -n    GS-score normalization option                            " << "\n"
  << "             1: normalize using larger glycan                      " << "\n"
  << "             2: normalize using smaller glycan                     " << "\n"
  << "             3: normalize using s1 structure                       " << "\n"
  << "    -o    output option                                            " << "\n"
  << "             1: score only                                         " << "\n"
  << "             2: score, superposed s2 structure with matrix, and sequence alignment" << "\n"
  << "    -help                                                          " << "\n"
  << "\n"
  << "outputs:                                                           " << "\n"
  << "     GS-score      : calculated using all aligned residue pairs" << "\n"
  << "     RMSD (Å)      : calculated using aligned residue pairs within 5 Å" << "\n"
  << "     ali_struct.pdb: the PDB coordinates of s2 structure aligned onto s1 structure" << "\n"
  << "     matrix.txt    : translational and rotational matrix to align s2 structure onto s1 structure" << "\n"
  << "     Seq. Alignment: a list of s2 residues aligned onto s1 residues within 5 Å" << "\n";


  exit(1);
}


void parseCommandLine( int argc, char *argv[] ){
  string a;
  progname = *argv;

  if( argc == 1 ) help( "" );
  
  for( ++argv ; --argc ; ++argv ){
    a = *argv;

    if( a == "-s1" ){
      if( --argc == 0 ) help( "ERROR: " + a + " requires value." );
      s1 = *++argv;
    } 	
	else if( a == "-s2" ){
      if( --argc == 0 ) help( "ERROR: " + a + " requires value." );
      s2 = *++argv;
    }  
    else if( a == "-n" ){
      if( --argc == 0 ) help( "ERROR: " + a + " requires value." );
      N_target_option = (int)atof( *++argv );
      
      if( N_target_option < 1 ) help( "ERROR: -n value" );
      if( N_target_option > 3 ) help( "ERROR: -n value" );
    } 
    else if( a == "-o" ){
      if( --argc == 0 ) help( "ERROR: " + a + " requires value." );
      output_option = (int)atof( *++argv );
      
      if( output_option < 1 ) help( "ERROR: -o value" );
      if( output_option > 2 ) help( "ERROR: -o value" );
    }    
	else if (a=="-help") help( "" );
    else help( "ERROR: Bad option: " + a );
  }  
}







// ************************************************************************
// *********** Class and Functions for Maximum Clique Detection ***********
// ************************************************************************
// ===============================================================
// There are a class and functions to identify maximum clique based on an improved
// branch and bound algorithm
// Reference:
// Janez Konc and Dusanka Janezic. An improved branch and bound algorithm for the 
// maximum clique problem. MATCH Commun. Math. Comput. Chem., 2007, 58, 569-590.
//
// Modified by H. S. Lee
// ==============================================================

class Maxclique {
	const bool* const* e;
	int pk, level;
	const float Tlimit;
  
	class Vertices {
		class Vertex {
  	    	int i, d;
    	
    		public:
      			void set_i( const int ii ) { i = ii; }
      			int get_i() const { return i; }
      			void set_degree( int dd ) { d = dd; }
      			int get_degree() const { return d; }
      	};
    
    	Vertex *v;
    	int sz;
    	static bool desc_degree( const Vertex vi, const Vertex vj) { return (vi.get_degree() > vj.get_degree()); }
  		
  		public:
    		Vertices( int size ) : sz(0) { v = new Vertex[size]; }
    		~Vertices () {}
    		void dispose() { if (v) delete [] v; }
    		void sort() { std::sort( v, v+sz, desc_degree ); }
    		void init_colors();
    		void set_degrees( Maxclique& );
    		int size() const { return sz; }
    		void push( const int ii ) { v[sz++].set_i(ii); };
   			void pop() { sz--; };
    		Vertex& at( const int ii ) const { return v[ii]; };
    		Vertex& end() const { return v[sz - 1]; };
	};
  
	class ColorClass {
    	int *i;
    	int sz;
  
  		public:
    		ColorClass() : sz(0), i(0) {}
    		ColorClass( const int sz ) : sz(sz), i(0) { init(sz); }
    		~ColorClass() { if (i) delete [] i; }
    		void init( const int sz ) { i = new int[sz]; rewind(); }
    		void push( const int ii ) { i[sz++] = ii; };
    		void pop() { sz--; };
   			void rewind() { sz = 0; };
    		int size() const { return sz; }
    		int& at(const int ii) const { return i[ii]; }
    		ColorClass& operator=( const ColorClass& dh ) {
	      		for (int j = 0; j < dh.sz; j++) i[j] = dh.i[j];      		
    	  		sz = dh.sz;      		
      			return *this;
			}
	};
	
	Vertices V;
	ColorClass *C, QMAX, Q;
	
	class StepCount {
    	int i1, i2;
  
  		public:
    		StepCount() : i1(0), i2(0) {}
    		void set_i1( const int ii )  { i1 = ii; }
    		int get_i1() const { return i1; }
   			void set_i2( const int ii )  { i2 = ii; }
   			int get_i2() const { return i2; }
    		void inc_i1()  { i1++; }
	};
  
	StepCount *S;
  	bool connection(const int i, const int j ) const { return e[i][j]; }
  	bool cut1( const int, const ColorClass& );
 	void cut2( const Vertices&, Vertices& );
  	void color_sort( Vertices& );
  	void expand( Vertices );
  	void expand_dyn( Vertices );
  	void _mcq( int*&, int&, bool );
  	void degree_sort( Vertices &R ) { R.set_degrees(*this); R.sort(); }
	
	public:
		Maxclique( const bool* const*, const int, const float=0.025 );
  		int steps() const { return pk; }
  		void mcq( int* &maxclique, int &sz ) { _mcq( maxclique, sz, false ); }
  		void mcqdyn( int* &maxclique, int &sz ) { _mcq(maxclique, sz, true); }
  		~Maxclique() {
    		if( C ) delete [] C;
    		if( S ) delete [] S;
   			V.dispose();
  	};
};



Maxclique::Maxclique ( const bool* const* conn, const int sz, const float tt) : pk(0), level(1), Tlimit(tt), V(sz), Q(sz), QMAX(sz ) {
	assert(conn!=0 && sz>0);
  	
  	for (int i=0; i < sz; i++) V.push(i);
  	
  	e = conn;
  	C = new ColorClass[sz + 1];
  	
  	for(int i=0; i < sz + 1; i++) C[i].init(sz + 1);
  	
  	S = new StepCount[sz + 1];
}



void Maxclique::_mcq( int* &maxclique, int &sz, bool dyn ) { 
	V.set_degrees(*this);
  	V.sort();
  	V.init_colors();
  
  	if( dyn ) {
    	for ( int i=0; i < V.size() + 1; i++ ) {
      		S[i].set_i1(0);
      		S[i].set_i2(0);
    	}
   
		expand_dyn(V);
	}
  	else
    	expand(V);
  	
  	maxclique = new int[QMAX.size()]; 
  	
  	for( int i=0; i<QMAX.size(); i++ ) { 
    	maxclique[i] = QMAX.at(i);
  	}
  	
  	sz = QMAX.size();
}



void Maxclique::Vertices::init_colors() { 
	const int max_degree = v[0].get_degree();
  
  	for( int i = 0; i < max_degree; i++ )
   		v[i].set_degree(i + 1);
  
  	for( int i = max_degree; i < sz; i++ )
    	v[i].set_degree(max_degree + 1);
}



void Maxclique::Vertices::set_degrees( Maxclique &m ) { 
  for( int i=0; i < sz; i++ ) {
    int d = 0;
    
    for( int j=0; j < sz; j++ )
    	if( m.connection(v[i].get_i(), v[j].get_i()) ) d++;
    
    	v[i].set_degree(d);
	}
}



bool Maxclique::cut1( const int pi, const ColorClass &A ) {
	for( int i = 0; i < A.size(); i++ )
    	if( connection(pi, A.at(i)) ) return true;
  
  	return false;
}



void Maxclique::cut2( const Vertices &A, Vertices &B ) {
	for( int i = 0; i < A.size() - 1; i++ ) {
    	if( connection(A.end().get_i(), A.at(i).get_i()) ) 
    		B.push(A.at(i).get_i());
	}
}



void Maxclique::color_sort( Vertices &R ) {
	int j = 0;
  	int maxno = 1;
  	int min_k = QMAX.size() - Q.size() + 1;
  	C[1].rewind();
  	C[2].rewind();
 	int k = 1;
 	
 	for( int i=0; i < R.size(); i++ ) {
    	int pi = R.at(i).get_i();
   		k = 1;
    	
    	while( cut1(pi, C[k]) )
      		k++;
    
    	if( k > maxno ) {
      		maxno = k;
      		C[maxno + 1].rewind();
    	}
    
    	C[k].push(pi);
    
    	if( k < min_k ) {
      		R.at(j++).set_i(pi);
    	}
  	}
  
  	if( j > 0 ) R.at(j-1).set_degree(0);
  	if( min_k <= 0 ) min_k = 1;
 	
 	for( k = min_k; k <= maxno; k++ ) {
    	for( int i = 0; i < C[k].size(); i++ ) {
      		R.at(j).set_i(C[k].at(i));
      		R.at(j++).set_degree(k);
    	}
    }
}



void Maxclique::expand( Vertices R ) {
	while( R.size() ) {
    	if( Q.size() + R.end().get_degree() > QMAX.size() ) {
      		Q.push(R.end().get_i());
      		Vertices Rp(R.size());
      		cut2(R, Rp);
      		
      		if( Rp.size() ) {
        		color_sort(Rp);
				pk++;
        		expand(Rp);
      		}
      		else if( Q.size() > QMAX.size() ) { 
				QMAX = Q;
      		}    
      		
      		Rp.dispose();
      		Q.pop();
    	}
    	else {
      		return;
    	}
    	
    	R.pop();
  	}
}



void Maxclique::expand_dyn( Vertices R ) {
	S[level].set_i1(S[level].get_i1() + S[level - 1].get_i1() - S[level].get_i2());
  	S[level].set_i2(S[level - 1].get_i1());
  
  	while( R.size() ) {
    	if( Q.size() + R.end().get_degree() > QMAX.size() ) {
      		Q.push(R.end().get_i());
     		Vertices Rp(R.size());
      		cut2(R, Rp);
      
      		if( Rp.size() ) {
        		if( (float)S[level].get_i1()/++pk < Tlimit ) {
          			degree_sort(Rp);
        		}
        		
        		color_sort(Rp);
				S[level].inc_i1();
				level++;
				expand_dyn(Rp);
				level--;
      		}
      		else if( Q.size() > QMAX.size() ) { 
				QMAX = Q;
      		}    
     
     		Rp.dispose();
      		Q.pop();
    	}
    	else{
      		return;
    	}
    
    	R.pop();
  	}
}
// *************************************************




int getNumberOfResidues( string filename ){
	char buffer[100];
	int max_num = 0;

	FILE *pInputFile = fopen( filename.c_str(), "r" );

	char resSeq[5];
	int res_num = 0;
	
	while( fgets( buffer, 100, pInputFile ) != NULL ){	
		if( buffer[0] == 'T' && buffer[1] == 'E' ){	
			break;
		}
		else if( ( buffer[0] == 'A' && buffer[1] == 'T' ) || ( buffer[0] == 'H' && buffer[1] == 'E' ) ){
			for( int i = 0 ; i < 4 ; i++ )
				resSeq[i] = buffer[22+i]; 
		
			resSeq[4] = '\0';	
		
			res_num = (int)atof( resSeq );		
			
			if( res_num > max_num )
				max_num = res_num;		
		}		
	}

	fclose( pInputFile );
	
	return max_num;
}




void readOGAndRingCentroidFromPDBFile( string filename, vector< Atom * > &vStruct ){
	char buffer[100];

	Atom *atomRing;
	Atom *atomOG;
	char X_char[9], Y_char[9], Z_char[9];

	FILE *pInputFile = fopen( filename.c_str(), "r" );

	double sum_ring_X = 0;
	double sum_ring_Y = 0;
	double sum_ring_Z = 0;
	
	char resSeq[5];
	int prev_res_num = 1;
	int curr_res_num = 0;
	
	atomRing = new Atom();
	
	while( fgets( buffer, 100, pInputFile ) != NULL ){	
		if( buffer[0] == 'T' && buffer[1] == 'E' ){			
			atomRing->R[0] = sum_ring_X/6.0;
			atomRing->R[1] = sum_ring_Y/6.0;
			atomRing->R[2] = sum_ring_Z/6.0;	
			
			atomRing->resSeq = curr_res_num;	
			atomRing->atom_name = 'R';						

			vStruct.push_back( atomRing );	

			break;
		}
		else if( ( buffer[0] == 'A' && buffer[1] == 'T' ) || ( buffer[0] == 'H' && buffer[1] == 'E' ) ){
			for( int i = 0 ; i < 4 ; i++ )
				resSeq[i] = buffer[22+i]; 
		
			resSeq[4] = '\0';	
		
			curr_res_num = (int)atof( resSeq );				
		
			if( curr_res_num != prev_res_num ){	
				atomRing->R[0] = sum_ring_X/6.0;
				atomRing->R[1] = sum_ring_Y/6.0;
				atomRing->R[2] = sum_ring_Z/6.0;	
				
				atomRing->resSeq = prev_res_num;
				atomRing->atom_name = 'R';	
				
				vStruct.push_back( atomRing );
					
				sum_ring_X = 0;
				sum_ring_Y = 0;
				sum_ring_Z = 0;	
				
				atomRing = new Atom();
				
				prev_res_num = curr_res_num;		
			}			
			
			if( buffer[13] == 'O' && buffer[14] == 'G' ){				
				atomOG = new Atom();			
				atomOG->atom_name = 'G';
					
				// Coordinates
				for( int i = 0 ; i < 8 ; i++ ){
					X_char[i] = buffer[30+i];
					Y_char[i] = buffer[38+i];
					Z_char[i] = buffer[46+i];
				}

				X_char[8] = '\0';
				Y_char[8] = '\0';
				Z_char[8] = '\0';
					
				atomOG->R[0] = (double)atof( X_char );
				atomOG->R[1] = (double)atof( Y_char );
				atomOG->R[2] = (double)atof( Z_char );	
				
				atomOG->resSeq = curr_res_num;													

				vStruct.push_back( atomOG );								
			}			
			else if( !( buffer[18] == 'N' && buffer[19] == 'E' && buffer[20] == '5' ) && 
					( ( buffer[13] == 'C' && buffer[14] == '1' ) || ( buffer[13] == 'C' && buffer[14] == '2' ) || ( buffer[13] == 'C' && buffer[14] == '3' ) || 
			         ( buffer[13] == 'C' && buffer[14] == '4' ) || ( buffer[13] == 'C' && buffer[14] == '5' ) || ( buffer[13] == 'O' && buffer[14] == '5' ) ) ){
			    // Coordinates
				for( int i = 0 ; i < 8 ; i++ ){
					X_char[i] = buffer[30+i];
					Y_char[i] = buffer[38+i];
					Z_char[i] = buffer[46+i];
				}

				X_char[8] = '\0';
				Y_char[8] = '\0';
				Z_char[8] = '\0';
					
				sum_ring_X += (double)atof( X_char );
				sum_ring_Y += (double)atof( Y_char );
				sum_ring_Z += (double)atof( Z_char );	
			}	
			else if( ( buffer[18] == 'N' && buffer[19] == 'E' && buffer[20] == '5' ) && 
				   ( ( buffer[13] == 'C' && buffer[14] == '2' ) || ( buffer[13] == 'C' && buffer[14] == '3' ) || ( buffer[13] == 'C' && buffer[14] == '4' ) || 
			         ( buffer[13] == 'C' && buffer[14] == '5' ) || ( buffer[13] == 'C' && buffer[14] == '6' ) || ( buffer[13] == 'O' && buffer[14] == '6' ) ) ){
			    // Coordinates
				for( int i = 0 ; i < 8 ; i++ ){
					X_char[i] = buffer[30+i];
					Y_char[i] = buffer[38+i];
					Z_char[i] = buffer[46+i];
				}

				X_char[8] = '\0';
				Y_char[8] = '\0';
				Z_char[8] = '\0';
					
				sum_ring_X += (double)atof( X_char );
				sum_ring_Y += (double)atof( Y_char );
				sum_ring_Z += (double)atof( Z_char );	
			}			
		}		
	}

	fclose( pInputFile );
}




void readStructureForScoring( string filename, vector< Residue * > &vStruct ){
	char buffer[100];
	
	Residue *res;
	char X_char[9], Y_char[9], Z_char[9];

	FILE *pInputFile = fopen( filename.c_str(), "r" );

	double sum_ring_X = 0;
	double sum_ring_Y = 0;
	double sum_ring_Z = 0;
	
	char resSeq[5];
	int prev_res_num = 1;
	int curr_res_num = 0;
	
	int num_atom = 0;
	char prev_resName[4];
	char curr_resName[4];
	
	res = new Residue();
	
	while( fgets( buffer, 100, pInputFile ) != NULL ){	
		if( buffer[0] == 'T' && buffer[1] == 'E' ){			
			res->R[0] = sum_ring_X/6.0;
			res->R[1] = sum_ring_Y/6.0;
			res->R[2] = sum_ring_Z/6.0;	
			
			res->resSeq = curr_res_num;	
			res->hasOG = 0;
			
			for( int i = 0 ; i < 4 ; i++ )
				res->resName[i] = curr_resName[i]; 
			
			vStruct.push_back( res );	

			break;
		}
		else if( ( buffer[0] == 'A' && buffer[1] == 'T' ) || ( buffer[0] == 'H' && buffer[1] == 'E' ) ){
			num_atom++;
			
			if( num_atom == 1 ){
				for( int i = 0 ; i < 3 ; i++ )
					prev_resName[i] = buffer[17+i]; 
				
				prev_resName[3] = '\0';	
			}
			
			for( int i = 0 ; i < 4 ; i++ )
				resSeq[i] = buffer[22+i]; 
		
			resSeq[4] = '\0';	
			
			curr_res_num = (int)atof( resSeq );		
			
			for( int i = 0 ; i < 3 ; i++ )
				curr_resName[i] = buffer[17+i]; 
		
			curr_resName[3] = '\0';			
		
			if( curr_res_num != prev_res_num ){	
				res->R[0] = sum_ring_X/6.0;
				res->R[1] = sum_ring_Y/6.0;
				res->R[2] = sum_ring_Z/6.0;	
				
				res->resSeq = prev_res_num;								
				res->hasOG = 0;	
				
				for( int i = 0 ; i < 4 ; i++ )
					res->resName[i] = prev_resName[i]; 
				
				vStruct.push_back( res );
					
				sum_ring_X = 0;
				sum_ring_Y = 0;
				sum_ring_Z = 0;	
				
				res = new Residue();
				
				prev_res_num = curr_res_num;
				
				for( int i = 0 ; i < 4 ; i++ )
					prev_resName[i] = curr_resName[i]; 		
			}			
			
			if( !( buffer[18] == 'N' && buffer[19] == 'E' && buffer[20] == '5' ) && 
			   ( ( buffer[13] == 'C' && buffer[14] == '1' ) || ( buffer[13] == 'C' && buffer[14] == '2' ) || ( buffer[13] == 'C' && buffer[14] == '3' ) || 
			     ( buffer[13] == 'C' && buffer[14] == '4' ) || ( buffer[13] == 'C' && buffer[14] == '5' ) || ( buffer[13] == 'O' && buffer[14] == '5' ) ) ){
			    // Coordinates
				for( int i = 0 ; i < 8 ; i++ ){
					X_char[i] = buffer[30+i];
					Y_char[i] = buffer[38+i];
					Z_char[i] = buffer[46+i];
				}

				X_char[8] = '\0';
				Y_char[8] = '\0';
				Z_char[8] = '\0';
					
				sum_ring_X += (double)atof( X_char );
				sum_ring_Y += (double)atof( Y_char );
				sum_ring_Z += (double)atof( Z_char );	
				
				if( buffer[13] == 'C' && buffer[14] == '1' ){
					res->Ring[0][0] = (double)atof( X_char );
					res->Ring[0][1] = (double)atof( Y_char );
					res->Ring[0][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '2' ){
					res->Ring[1][0] = (double)atof( X_char );
					res->Ring[1][1] = (double)atof( Y_char );
					res->Ring[1][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '3' ){
					res->Ring[2][0] = (double)atof( X_char );
					res->Ring[2][1] = (double)atof( Y_char );
					res->Ring[2][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '4' ){
					res->Ring[3][0] = (double)atof( X_char );
					res->Ring[3][1] = (double)atof( Y_char );
					res->Ring[3][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '5' ){
					res->Ring[4][0] = (double)atof( X_char );
					res->Ring[4][1] = (double)atof( Y_char );
					res->Ring[4][2] = (double)atof( Z_char );
				}
				else{
					res->Ring[5][0] = (double)atof( X_char );
					res->Ring[5][1] = (double)atof( Y_char );
					res->Ring[5][2] = (double)atof( Z_char );
				}
			}	
			else if( ( buffer[18] == 'N' && buffer[19] == 'E' && buffer[20] == '5' ) && 
			       ( ( buffer[13] == 'C' && buffer[14] == '2' ) || ( buffer[13] == 'C' && buffer[14] == '3' ) || ( buffer[13] == 'C' && buffer[14] == '4' ) || 
			         ( buffer[13] == 'C' && buffer[14] == '5' ) || ( buffer[13] == 'C' && buffer[14] == '6' ) || ( buffer[13] == 'O' && buffer[14] == '6' ) ) ){
			    // Coordinates
				for( int i = 0 ; i < 8 ; i++ ){
					X_char[i] = buffer[30+i];
					Y_char[i] = buffer[38+i];
					Z_char[i] = buffer[46+i];
				}

				X_char[8] = '\0';
				Y_char[8] = '\0';
				Z_char[8] = '\0';
					
				sum_ring_X += (double)atof( X_char );
				sum_ring_Y += (double)atof( Y_char );
				sum_ring_Z += (double)atof( Z_char );	
				
				if( buffer[13] == 'C' && buffer[14] == '2' ){
					res->Ring[0][0] = (double)atof( X_char );
					res->Ring[0][1] = (double)atof( Y_char );
					res->Ring[0][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '3' ){
					res->Ring[1][0] = (double)atof( X_char );
					res->Ring[1][1] = (double)atof( Y_char );
					res->Ring[1][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '4' ){
					res->Ring[2][0] = (double)atof( X_char );
					res->Ring[2][1] = (double)atof( Y_char );
					res->Ring[2][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '5' ){
					res->Ring[3][0] = (double)atof( X_char );
					res->Ring[3][1] = (double)atof( Y_char );
					res->Ring[3][2] = (double)atof( Z_char );
				}
				else if( buffer[13] == 'C' && buffer[14] == '6' ){
					res->Ring[4][0] = (double)atof( X_char );
					res->Ring[4][1] = (double)atof( Y_char );
					res->Ring[4][2] = (double)atof( Z_char );
				}
				else{
					res->Ring[5][0] = (double)atof( X_char );
					res->Ring[5][1] = (double)atof( Y_char );
					res->Ring[5][2] = (double)atof( Z_char );
				}
			}			
		}		
	}

	fclose( pInputFile );
}




void setGlycosidicOxygens( string filename, vector< Residue * > &vStruct ){
	FILE *pInputFile = fopen( filename.c_str(), "r" );	
	
	char buffer[100];	
	char X_char[9], Y_char[9], Z_char[9];
	char resSeq[5];
	char t_res[3];
	int index;
	int res_num;
	
	while( fgets( buffer, 100, pInputFile ) != NULL ){	
		if( buffer[0] == 'T' && buffer[1] == 'E' ){				
			break;
		}
		else if( ( buffer[0] == 'A' && buffer[1] == 'T' ) || ( buffer[0] == 'H' && buffer[1] == 'E' ) ){
			if( buffer[13] == 'O' && buffer[14] == 'G' ){
				t_res[0] = buffer[15];
				t_res[1] = buffer[16];
				t_res[2] = '\0';
				
				index = (int)atof( t_res );
				
				for( int i = 0 ; i < 4 ; i++ )
					resSeq[i] = buffer[22+i]; 
		
				resSeq[4] = '\0';	
		
				res_num = (int)atof( resSeq );		
				
				for( int i = 0 ; i < 8 ; i++ ){
					X_char[i] = buffer[30+i];
					Y_char[i] = buffer[38+i];
					Z_char[i] = buffer[46+i];
				}

				X_char[8] = '\0';
				Y_char[8] = '\0';
				Z_char[8] = '\0';
				
				for( int i = 0 ; i < vStruct.size() ; i++ ){
					if( vStruct[i]->resSeq == index ){
						vStruct[i]->OG[0] = (double)atof( X_char );
						vStruct[i]->OG[1] = (double)atof( Y_char );
						vStruct[i]->OG[2] = (double)atof( Z_char );
						vStruct[i]->hasOG = 1;
						
						vStruct[i]->resSeq_OG_linked = res_num;
						
						break;
					}
				}
			}
		}		
	}

	fclose( pInputFile );
}




void GeneratePairs( vector< Atom * > &vRefStruct, vector< Atom * > &vFitStruct, vector< Pair * > &vPair ){
	int num_element_ref = vRefStruct.size();
	int num_element_fit = vFitStruct.size();

	Pair *p; 

	for( int i = 0 ; i < num_element_ref ; i++ ){
		for( int j = 0 ; j < num_element_fit ; j++ ){
			if( vRefStruct[i]->atom_name == vFitStruct[j]->atom_name ){
				p = new Pair();
			    p->ref_index = i;
				p->fit_index = j;

				vPair.push_back( p );
			}
		}
	}
}




void GenerateEdges( vector< Atom * > &vRefStruct, vector< Atom * > &vFitStruct, vector< Pair * > &vPair, vector< Edge * > &vEdge, float dist_tol_cutoff ){
	Edge *e; 
	int num_pairs = vPair.size();
	double dist_1, dist_2, d_dist;	

	for( int i = 0 ; i < num_pairs ; i++ ){
		for( int j = i+1 ; j < num_pairs ; j++ ){
			dist_1 = sqrt( pow( vRefStruct[vPair[i]->ref_index]->R[0] - vRefStruct[vPair[j]->ref_index]->R[0], 2 ) 
				+ pow( vRefStruct[vPair[i]->ref_index]->R[1] - vRefStruct[vPair[j]->ref_index]->R[1], 2 ) 
				+ pow( vRefStruct[vPair[i]->ref_index]->R[2] - vRefStruct[vPair[j]->ref_index]->R[2], 2 ) );

			dist_2 = sqrt( pow( vFitStruct[vPair[i]->fit_index]->R[0] - vFitStruct[vPair[j]->fit_index]->R[0], 2 ) 
				+ pow( vFitStruct[vPair[i]->fit_index]->R[1] - vFitStruct[vPair[j]->fit_index]->R[1], 2 ) 
				+ pow( vFitStruct[vPair[i]->fit_index]->R[2] - vFitStruct[vPair[j]->fit_index]->R[2], 2 ) );

			d_dist = dist_1 - dist_2;

			if( d_dist < dist_tol_cutoff && d_dist > -dist_tol_cutoff ){     
				e = new Edge();
				e->pair_i = i;
				e->pair_j = j;

				vEdge.push_back( e );
			}
		}
	}
}




void writePairs( vector< Pair * > &vPair ){
	int num_pairs = vPair.size();
	
	FILE *pOutputFile_pair = fopen( "pairs.rst", "w" );

	fprintf( pOutputFile_pair, "%s", "pair_index\tref_index\tfit_index\n" );

	for( int i = 0 ; i < num_pairs ; i++ ){
		fprintf( pOutputFile_pair, "%d", i );
		fprintf( pOutputFile_pair, "%s", "\t" );
		fprintf( pOutputFile_pair, "%d", vPair[i]->ref_index );
		fprintf( pOutputFile_pair, "%s", "\t" );
		fprintf( pOutputFile_pair, "%d", vPair[i]->fit_index );
		fprintf( pOutputFile_pair, "\n" );
	}

	fprintf( pOutputFile_pair, "END\n" );

	fclose( pOutputFile_pair );
}




void writeProductGraph( int num_pairs, vector< Edge * > &vEdge ){
	FILE *pOutputFile = fopen( "product_graph.rst", "w" );	

	fprintf( pOutputFile, "%s", "p edge " );
	fprintf( pOutputFile, "%d", num_pairs );
	fprintf( pOutputFile, "%s", " P=0.5 SEED=12415\n" );
	
	int num_edges = vEdge.size();

	for( int i = 0 ; i < num_edges ; i++ ){
		fprintf( pOutputFile, "%s", "e " );

		if( vEdge[i]->pair_i < vEdge[i]->pair_j ){
			fprintf( pOutputFile, "%d", vEdge[i]->pair_i );
			fprintf( pOutputFile, "%s", " " );
			fprintf( pOutputFile, "%d", vEdge[i]->pair_j );
			fprintf( pOutputFile, "\n" );
		}
		else{
			fprintf( pOutputFile, "%d", vEdge[i]->pair_j );
			fprintf( pOutputFile, "%s", " " );
			fprintf( pOutputFile, "%d", vEdge[i]->pair_i );
			fprintf( pOutputFile, "\n" );
		}
	}

	fclose( pOutputFile );
}




void read_dimacs( string name, bool** &conn, int &size ){
	ifstream f (name.c_str());
    string buffer;
    assert( f.is_open() );
    set<int> v;
    multimap<int,int> e;
    
    while( !getline( f, buffer ).eof() ){
    	if (buffer[0] == 'e') {
      		int vi, vj;
      		sscanf( buffer.c_str(), "%*c %d %d", &vi, &vj );
      		v.insert(vi);
      		v.insert(vj);
      		e.insert( make_pair( vi, vj ) );
    	}
  	}
  
  	size = *v.rbegin() + 1;
  	conn = new bool*[size];
  
  	for (int i=0; i < size; i++) {
    	conn[i] = new bool[size];
    	memset(conn[i], 0, size * sizeof(bool));
  	}
  
  	for (multimap<int,int>::iterator it = e.begin(); it != e.end(); it++) {
    	conn[it->first][it->second] = true;
    	conn[it->second][it->first] = true;
  	}
  
  	//cout << "|E| = " << e.size() << "  |V| = " << v.size() << " p = " << (double) e.size() / (v.size() * (v.size() - 1) / 2) << endl;
  
  	f.close();
}




int identifyMaxClique( vector< int > &vPairs ){
	bool **conn;
    int size;		

  	read_dimacs( "product_graph.rst", conn, size );
  	
  	clock_t start1, start2;
  	int *qmax;
  	int qsize;

	start1 = time(NULL);
    start2 = clock();
    
    Maxclique md( conn, size, 0.025 );
    md.mcqdyn( qmax, qsize ); 
   
    for( int i = 0; i < qsize; i++ ){
    	vPairs.push_back( qmax[i] );
    }
    	
    delete [] qmax;
    
    for (int i=0 ; i < size ; i++ )
      delete [] conn[i];
      
    delete [] conn;
    
        	
    if( qsize < 3 ){
    	return 0;
    }
    
    
    return 1;
}




void superposeFitStructure( string filename, string output_name ){
	char buffer[100];

	double r[3];
	double r_prime[3];

	char X_char[9], Y_char[9], Z_char[9];

	FILE *pInputFile = fopen( filename.c_str(), "r" );
	FILE *pOutputFile = fopen( output_name.c_str(), "w" );

	while( 1 ){
		fgets( buffer, 100, pInputFile );

		if( buffer[0] == 'T' && buffer[1] == 'E' )
			break;

		if( ( buffer[0] == 'A' && buffer[1] == 'T' ) || ( buffer[0] == 'H' && buffer[1] == 'E' ) ){
			for( int i = 0 ; i < 30 ; i++ )
				fprintf( pOutputFile, "%c", buffer[i] );

			for( int i = 0 ; i < 8 ; i++ ){
				X_char[i] = buffer[30+i];
				Y_char[i] = buffer[38+i];
				Z_char[i] = buffer[46+i];
			}

			X_char[8] = '\0';
			Y_char[8] = '\0';
			Z_char[8] = '\0';

			r[0] = (double)atof( X_char );
			r[1] = (double)atof( Y_char );
			r[2] = (double)atof( Z_char );

			r_prime[0] = t[0] + u[0][0]*r[0] + u[0][1]*r[1] + u[0][2]*r[2];
			r_prime[1] = t[1] + u[1][0]*r[0] + u[1][1]*r[1] + u[1][2]*r[2];
			r_prime[2] = t[2] + u[2][0]*r[0] + u[2][1]*r[1] + u[2][2]*r[2];		
			
			fprintf( pOutputFile, "%8.3f", r_prime[0] );
			fprintf( pOutputFile, "%8.3f", r_prime[1] );
			fprintf( pOutputFile, "%8.3f", r_prime[2] );

			fprintf( pOutputFile, "\n" );
		}		
	}

	fprintf( pOutputFile, "TER" );

	fclose( pOutputFile );
	fclose( pInputFile );	
}




void setSuperposedCoordinatesForFitStructure( vector< Residue * > &vFitStruct ){
	double r[3];
	double r_prime[3];
	int num_residues = vFitStruct.size();
	
	for( int i = 0 ; i < num_residues ; i++ ){
		// Ring atoms
		for( int j = 0 ; j < 6 ; j++ ){
			r[0] = vFitStruct[i]->Ring[j][0];
			r[1] = vFitStruct[i]->Ring[j][1];
			r[2] = vFitStruct[i]->Ring[j][2];
			
			r_prime[0] = t[0] + u[0][0]*r[0] + u[0][1]*r[1] + u[0][2]*r[2];
			r_prime[1] = t[1] + u[1][0]*r[0] + u[1][1]*r[1] + u[1][2]*r[2];
			r_prime[2] = t[2] + u[2][0]*r[0] + u[2][1]*r[1] + u[2][2]*r[2];	
			
			vFitStruct[i]->Ring_sup[j][0] = r_prime[0];
			vFitStruct[i]->Ring_sup[j][1] = r_prime[1];
			vFitStruct[i]->Ring_sup[j][2] = r_prime[2];		
		}
		
		// Ring centroid
		r[0] = vFitStruct[i]->R[0];
		r[1] = vFitStruct[i]->R[1];
		r[2] = vFitStruct[i]->R[2];
			
		r_prime[0] = t[0] + u[0][0]*r[0] + u[0][1]*r[1] + u[0][2]*r[2];
		r_prime[1] = t[1] + u[1][0]*r[0] + u[1][1]*r[1] + u[1][2]*r[2];
		r_prime[2] = t[2] + u[2][0]*r[0] + u[2][1]*r[1] + u[2][2]*r[2];	
			
		vFitStruct[i]->R_sup[0] = r_prime[0];
		vFitStruct[i]->R_sup[1] = r_prime[1];
		vFitStruct[i]->R_sup[2] = r_prime[2];	
		
		// OG atom
		r[0] = vFitStruct[i]->OG[0];
		r[1] = vFitStruct[i]->OG[1];
		r[2] = vFitStruct[i]->OG[2];
			
		r_prime[0] = t[0] + u[0][0]*r[0] + u[0][1]*r[1] + u[0][2]*r[2];
		r_prime[1] = t[1] + u[1][0]*r[0] + u[1][1]*r[1] + u[1][2]*r[2];
		r_prime[2] = t[2] + u[2][0]*r[0] + u[2][1]*r[1] + u[2][2]*r[2];	
			
		vFitStruct[i]->OG_sup[0] = r_prime[0];
		vFitStruct[i]->OG_sup[1] = r_prime[1];
		vFitStruct[i]->OG_sup[2] = r_prime[2];		
	}
}




void setRotationMatrix( vector< Atom * > &vRefStruct, vector< Atom * > &vFitStruct ){
	int n = vRefStruct.size();    
	int mode = 1;
	int ier;

	int gj, gk, gl, gm1, gm;
	int ip[9] = { 0, 1, 3, 1, 2, 4, 3, 4, 5 };
	int ip1201[4] = { 1, 2, 0, 1 };

	double r[3][3], xc[3], yc[3], wc;
	double a[3][3], b[3][3], e[3], rr[6], ss[6];
	double e0, d, spur, det, cof, h, g;
	double cth, sth, sqrth, p, sigma;
	double sqrt3, tol;

	sqrt3 = 1.73205080756888;
    tol = 0.01;

	wc = 0.0;
    double rms = 0.0;
    e0 = 0.0;

    for( int i = 0 ; i < 3 ; i++ ){
		xc[i] = 0.0;
        yc[i] = 0.0;
        t[i] = 0.0;

		 for( int j = 0 ; j < 3 ; j++ ){
			 r[i][j] = 0.0;
             u[i][j] = 0.0;
             a[i][j] = 0.0;
			 
			 if( i == j ){
               u[i][j] = 1.0;
               a[i][j] = 1.0;
			 }
		 }
	}

	ier = -1;

	if( n <  1 ) 
		return;
    
	ier = -2;

	for( int m = 0 ; m < n ; m++ ){
		wc = wc + 1.0;

		for( int i = 0 ; i < 3 ; i++ ){
			xc[i] = xc[i] + vFitStruct[m]->R[i];
            yc[i] = yc[i] + vRefStruct[m]->R[i];
		}
	}

	if( wc <= 0 ) return;

	for( int i = 0 ; i < 3 ; i++ ){
		xc[i] = xc[i] / wc;
        yc[i] = yc[i] / wc;
	}

	for( int m = 0 ; m < n ; m++ ){
		for( int i = 0 ; i < 3 ; i++ ){
			e0 = e0 + ( pow( vFitStruct[m]->R[i] - xc[i], 2 ) + pow( vRefStruct[m]->R[i] - yc[i], 2 ) );
            d = vRefStruct[m]->R[i] - yc[i];

			for( int j = 0 ; j < 3 ; j++ ){
				r[i][j] = r[i][j] + d * ( vFitStruct[m]->R[j] - xc[j] );
			}
		}
	}

	det = r[0][0] * ( r[1][1] * r[2][2] - r[1][2] * r[2][1] ) 
		- r[0][1] * ( r[1][0] * r[2][2] - r[1][2] * r[2][0] )
		+ r[0][2] * ( r[1][0] * r[2][1] - r[1][1] * r[2][0] );

	sigma = det;

	int index = -1;

	for( int j = 0 ; j < 3 ; j++ ){
		for( int i = 0 ; i <= j ; i++ ){
			index++;
            rr[index] = r[0][i] * r[0][j] + r[1][i] * r[1][j] + r[2][i] * r[2][j];
		}
	}

	 spur = ( rr[0] + rr[2] + rr[5] ) / 3.0;

	 cof = ( rr[2]*rr[5] - rr[4]*rr[4] + rr[0]*rr[5] - rr[3]*rr[3] + rr[0]*rr[2] - rr[1]*rr[1] ) / 3.0;

	 det = det * det;

	 for( int i = 0 ; i < 3 ; i++ ){
		 e[i] = spur;
	 }	 	 

	 if( spur <= 0 ){
		 for( int i = 0 ; i < 3 ; i++ ){
			 t[i] = yc[i] - u[i][0] * xc[0] - u[i][1] * xc[1] - u[i][2] * xc[2];
		 }

		 for( int i = 0 ; i < 3 ; i++ ){
			 if( e[i] < 0 ) 
				 e[i] = 0.0;

			 e[i] = sqrt( e[i] );
		 }

		 ier = 0;

		 if( e[1] <= ( e[0] * 0.00005 ) ) 
			 ier = -1;

		 d = e[2];

		 if( sigma < 0.0 ){
			 d = -d;
			  
			 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
				 ier = -1;
		 }

		 d = ( d + e[1] ) + e[0];

		 rms = ( e0 - d ) - d;

		 if( rms < 0.0 ) 
			 rms = 0.0;

		 return;
	 }

	 d = spur * spur;
     h = d - cof;
     g = ( spur * cof - det ) / 2.0 - spur * h;

	 if( h <= 0 ){ 
		 if( mode == 0 ){
			 for( int i = 0 ; i < 3 ; i++ ){
				 if( e[i] < 0 ) 
					 e[i] = 0.0;

				 e[i] = sqrt( e[i] );
			 }

			 ier = 0;

             if( e[1] <= ( e[0] * 0.00005 ) ) 
				 ier = -1;

			 d = e[2];

			 if( sigma < 0.0 ){
				 d = -d;
				  
				 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
					 ier = -1;
			 }

			 d = ( d + e[1] ) + e[0];

			 rms = ( e0 - d ) - d;

			 if( rms < 0.0 ) 
				 rms = 0.0;

			 return;
		 }
		 else{
			 for( int l = 0 ; l < 2 ; l++ ){
				 d = 0.0;

				 for( int i = 0 ; i < 3 ; i++ ){
					 b[i][l] = r[i][0] * a[0][l] + r[i][1] * a[1][l] + r[i][2] * a[2][l];
					 d = d + pow( b[i][l], 2 );
				 }

				 if( d > 0 ) 
					 d = 1.0 / sqrt( d );

				 for( int i = 0 ; i < 3 ; i++ ){
					 b[i][l] = b[i][l] * d;
				 }
			 }

             d = b[0][0] * b[0][1] + b[1][0] * b[1][1] + b[2][0] * b[2][1];
             p = 0.0;

			 for( int i = 0 ; i < 3 ; i++ ){
				 b[i][1] = b[i][1] - d * b[i][0];
			     p = p + pow( b[i][1], 2 );
			 }

			 if( p <= tol ){
				  p = 1.0;

				  for( int i = 0 ; i < 3 ; i++ ){
					  if( p < abs( b[i][0] ) ){}
					  else{
						  p = abs( b[i][0] );
						  gj = i;
					  }
				  }

				  gk = ip1201[gj];
				  gl = ip1201[gj+1];
				  p = sqrt( pow( b[gk][0], 2 ) + pow( b[gl][0], 2 ) );

				  if( p <= tol ){
					 for( int i = 0 ; i < 3 ; i++ ){
						 t[i] = yc[i] - u[i][0] * xc[0] - u[i][1] * xc[1] - u[i][2] * xc[2];
					 }

					 for( int i = 0 ; i < 3 ; i++ ){
						 if( e[i] < 0 ) 
							 e[i] = 0.0;

						 e[i] = sqrt( e[i] );
					 }

					 ier = 0;

					 if( e[1] <= ( e[0] * 0.00005 ) ) 
						 ier = -1;

					 d = e[2];

					 if( sigma < 0.0 ){
						 d = -d;
						  
						 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
							 ier = -1;
					 }

					 d = ( d + e[1] ) + e[0];

					 rms = ( e0 - d ) - d;

					 if( rms < 0.0 ) 
						 rms = 0.0;

					 return;
				  }

				  b[gj][1] = 0.0;
                  b[gk][1] = -b[gl][0] / p;
                  b[gl][1] = b[gk][0] / p;
			 }
			 else{
				 p = 1.0 / sqrt( p );

				for( int i = 0 ; i < 3 ; i++ )
					b[i][1] = b[i][1] * p;
			 }

			 b[0][2] = b[1][0] * b[2][1] - b[1][1] * b[2][0];
             b[1][2] = b[2][0] * b[0][1] - b[2][1] * b[0][0];
             b[2][2] = b[0][0] * b[1][1] - b[0][1] * b[1][0];

			 for( int i = 0 ; i < 3 ; i++ ){
				 for( int j = 0 ; j < 3 ; j++ ){
					 u[i][j] = b[i][0] * a[j][0] + b[i][1] * a[j][1] + b[i][2] * a[j][2];
				 }
			 }

			 for( int i = 0 ; i < 3 ; i++ ){
				 t[i] = yc[i] - u[i][0] * xc[0] - u[i][1] * xc[1] - u[i][2] * xc[2];
			 }

			 for( int i = 0 ; i < 3 ; i++ ){
				 if( e[i] < 0 ) 
					 e[i] = 0.0;

				 e[i] = sqrt( e[i] );
			 }

			 ier = 0;

             if( e[1] <= ( e[0] * 0.00005 ) ) 
				 ier = -1;

			 d = e[2];

			 if( sigma < 0.0 ){
				 d = -d;
				  
				 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
					 ier = -1;
			 }

			 d = ( d + e[1] ) + e[0];

			 rms = ( e0 - d ) - d;

			 if( rms < 0.0 ) 
				 rms = 0.0;

			 return;
		 }
	 }
	 
     sqrth = sqrt( h );
	 d = h*h*h - g*g;

	 if( d < 0 ) d = 0.0;

	 d = atan2( sqrt(d), -g ) / 3.0;
     cth = sqrth * cos(d);
     sth = sqrth * sqrt3 * sin(d);
     e[0] = ( spur + cth ) + cth;
     e[1] = ( spur - cth ) + sth;
     e[2] = ( spur - cth ) - sth;

	 if( mode == 0 ){
		 for( int i = 0 ; i < 3 ; i++ ){
			 if( e[i] < 0 ) 
				 e[i] = 0.0;

			 e[i] = sqrt( e[i] );
		 }

		 ier = 0;

		 if( e[1] <= ( e[0] * 0.00005 ) ) 
			 ier = -1;

		 d = e[2];

		 if( sigma < 0.0 ){
			 d = -d;
			  
			 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
				 ier = -1;
		 }

		 d = ( d + e[1] ) + e[0];

		 rms = ( e0 - d ) - d;

		 if( rms < 0.0 ) 
			 rms = 0.0;

		 return;
	 }

     for( int l = 0 ; l < 3 ; l += 2 ){
		 d = e[l];
         ss[0] = ( d - rr[2] ) * ( d - rr[5] ) - rr[4] * rr[4];
         ss[1] = ( d - rr[5] ) * rr[1] + rr[3] * rr[4];
         ss[2] = ( d - rr[0] ) * ( d - rr[5] ) - rr[3] * rr[3];
         ss[3] = ( d - rr[2] ) * rr[3] + rr[1] * rr[4];
         ss[4] = ( d - rr[0] ) * rr[4] + rr[1] * rr[3];
         ss[5] = ( d - rr[0] ) * ( d - rr[2] ) - rr[1] * rr[1];

		 if( abs( ss[0] ) >= abs( ss[2] ) ){
			 gj = 0;

			 if( abs( ss[0] ) < abs( ss[5] ) )
				 gj = 2;
		 }
		 else if( abs( ss[2] ) >= abs( ss[5] ) )
			 gj = 1;
		 else
			 gj = 2;

		 d = 0.0;
         gj = 3 * gj;

		 for( int i = 0 ; i < 3 ; i++ ){
			 gk = ip[i+gj];
             a[i][l] = ss[gk];
             d = d + ss[gk] * ss[gk];
		 }

		 if( d > 0 )
			 d = 1.0 / sqrt(d);

		 for( int i = 0 ; i < 3 ; i++ ){
			 a[i][l] = a[i][l] * d;
		 }
	 }

	 d = a[0][0] * a[0][2] + a[1][0] * a[1][2] + a[2][0] * a[2][2];

	 if( ( e[0] - e[1] ) > ( e[1] - e[2] ) ){ 
         gm1 = 2;
         gm = 0;
	 }
	 else{
         gm1 = 0;
         gm = 2;
	 }

	 p = 0.0;

     for( int i = 0 ; i < 3 ; i++ ){
		 a[i][gm1] = a[i][gm1] - d * a[i][gm];
         p = p + pow( a[i][gm1], 2 );
	 }

	 if( p <= tol ){
		  p = 1.0;

		  for( int i = 0 ; i < 3 ; i++ ){
			  if( p < abs( a[i][gm] ) ){}
			  else{
				  p = abs( a[i][gm] );
			      gj = i;
			  }
		  }

		  gk = ip1201[gj];
          gl = ip1201[gj+1];
		  p = sqrt( pow( a[gk][gm], 2 ) + pow( a[gl][gm], 2 ) );

		  if( p <= tol ){
			 for( int i = 0 ; i < 3 ; i++ ){
				 t[i] = yc[i] - u[i][0] * xc[0] - u[i][1] * xc[1] - u[i][2] * xc[2];
			 }

			 for( int i = 0 ; i < 3 ; i++ ){
				 if( e[i] < 0 ) 
					 e[i] = 0.0;

				 e[i] = sqrt( e[i] );
			 }

			 ier = 0;

             if( e[1] <= ( e[0] * 0.00005 ) ) 
				 ier = -1;

			 d = e[2];

			 if( sigma < 0.0 ){
				 d = -d;
				  
				 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
					 ier = -1;
			 }

			 d = ( d + e[1] ) + e[0];

			 rms = ( e0 - d ) - d;

			 if( rms < 0.0 ) 
				 rms = 0.0;

			 return;
		  }

		  a[gj][gm1] = 0.0;
          a[gk][gm1] = -a[gl][gm] / p;
          a[gl][gm1] =  a[gk][gm] / p;
	 }
	 else{
		 p = 1.0 / sqrt(p);

		 for( int i = 0 ; i < 3 ; i++ ){
			 a[i][gm1] = a[i][gm1] * p;
		 }
	 }

	 a[0][1] = a[1][2] * a[2][0] - a[1][0] * a[2][2];
     a[1][1] = a[2][2] * a[0][0] - a[2][0] * a[0][2];
     a[2][1] = a[0][2] * a[1][0] - a[0][0] * a[1][2];

	 for( int l = 0 ; l < 2 ; l++ ){
		 d = 0.0;

		 for( int i = 0 ; i < 3 ; i++ ){
			 b[i][l] = r[i][0] * a[0][l] + r[i][1] * a[1][l] + r[i][2] * a[2][l];
			 d = d + pow( b[i][l], 2 );
		 }

		 if( d > 0 ) 
			 d = 1.0 / sqrt( d );

		 for( int i = 0 ; i < 3 ; i++ ){
			 b[i][l] = b[i][l] * d;
		 }
	 }

	 d = b[0][0] * b[0][1] + b[1][0] * b[1][1] + b[2][0] * b[2][1];
	 p = 0.0;

	 for( int i = 0 ; i < 3 ; i++ ){
		 b[i][1] = b[i][1] - d * b[i][0];
		 p = p + pow( b[i][1], 2 );
	 }

	 if( p <= tol ){
		  p = 1.0;

		  for( int i = 0 ; i < 3 ; i++ ){
			  if( p < abs( b[i][0] ) ){}
			  else{
				  p = abs( b[i][0] );
				  gj = i;
			  }
		  }

		  gk = ip1201[gj];
		  gl = ip1201[gj+1];
		  p = sqrt( pow( b[gk][0], 2 ) + pow( b[gl][0], 2 ) );

		  if( p <= tol ){
			 for( int i = 0 ; i < 3 ; i++ ){
				 t[i] = yc[i] - u[i][0] * xc[0] - u[i][1] * xc[1] - u[i][2] * xc[2];
			 }

			 for( int i = 0 ; i < 3 ; i++ ){
				 if( e[i] < 0 ) 
					 e[i] = 0.0;

				 e[i] = sqrt( e[i] );
			 }

			 ier = 0;

             if( e[1] <= ( e[0] * 0.00005 ) ) 
				 ier = -1;

			 d = e[2];

			 if( sigma < 0.0 ){
				 d = -d;
				  
				 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
					 ier = -1;
			 }

			 d = ( d + e[1] ) + e[0];

			 rms = ( e0 - d ) - d;

			 if( rms < 0.0 ) 
				 rms = 0.0;

			 return;
		  }

		  b[gj][1] = 0.0;
		  b[gk][1] = -b[gl][0] / p;
		  b[gl][1] = b[gk][0] / p;
	 }
	 else{
		 p = 1.0 / sqrt( p );

		 for( int i = 0 ; i < 3 ; i++ )
			b[i][1] = b[i][1] * p;
	 }

	 b[0][2] = b[1][0] * b[2][1] - b[1][1] * b[2][0];
	 b[1][2] = b[2][0] * b[0][1] - b[2][1] * b[0][0];
	 b[2][2] = b[0][0] * b[1][1] - b[0][1] * b[1][0];

	 for( int i = 0 ; i < 3 ; i++ ){
		 for( int j = 0 ; j < 3 ; j++ ){
			 u[i][j] = b[i][0] * a[j][0] + b[i][1] * a[j][1] + b[i][2] * a[j][2];
		 }
	 }

	 for( int i = 0 ; i < 3 ; i++ ){
		 t[i] = yc[i] - u[i][0] * xc[0] - u[i][1] * xc[1] - u[i][2] * xc[2];
	 }

	 for( int i = 0 ; i < 3 ; i++ ){
		 if( e[i] < 0 ) 
			 e[i] = 0.0;

		 e[i] = sqrt( e[i] );
	 }

	 ier = 0;

	 if( e[1] <= ( e[0] * 0.00005 ) ) 
		 ier = -1;

	 d = e[2];

	 if( sigma < 0.0 ){
		 d = -d;
		  
		 if( ( e[1] - e[2] ) <= ( e[0] * 0.00005 ) ) 
			 ier = -1;
	 }

	 d = ( d + e[1] ) + e[0];

	 rms = ( e0 - d ) - d;

	 if( rms < 0.0 ) 
		 rms = 0.0;

	 return;
}  




void setMaxCliqueAlignedNodes( vector< Atom * > &vRefStruct, vector< Atom * > &vFitStruct, vector< Atom * > &vRefAligned, vector< Atom * > &vFitAligned, vector< int > &vPairs ){
	char buffer[100];
	char index[5];
	
	FILE *pInputFile_pairs = fopen( "pairs.rst", "r" );
	int pair_index, res_ref, res_fit;
	int start_pos;

	fgets( buffer, 100, pInputFile_pairs );
	//cout << buffer;

	while( 1 ){
		fgets( buffer, 100, pInputFile_pairs );

		if( buffer[0] == 'E' )
			break;

		int pos = 0;

		for( int i = 0 ; i < 100 ; i++ ){
			if( buffer[i] == '\t' ){
				for( int j = 0 ; j < i ; j++ )
					index[j] = buffer[j];													
				
				index[i] = '\0';
				pair_index = (int)atof( index );
				start_pos = i+1;

				break;
			}		
		}

		for( int i = 0 ; i < vPairs.size() ; i++ ){
			if( pair_index == vPairs[i] ){
				for( int j = start_pos ; j < 100 ; j++ ){
					if( buffer[j] == '\t' ){
						int pos = 0;

						for( int k = start_pos ; k < j ; k++ ){
							index[pos] = buffer[k];		
							pos++;
						}
						
						index[pos] = '\0';
						res_ref = (int)atof( index );
						start_pos = j+1;

						break;
					}
				}		
				
				for( int j = start_pos ; j < 100 ; j++ ){
					if( buffer[j] == '\n' ){
						int pos = 0;

						for( int k = start_pos ; k < j ; k++ ){
							index[pos] = buffer[k];		
							pos++;
						}
						
						index[pos] = '\0';
						res_fit = (int)atof( index );

						break;
					}
				}				

				//cout << pair_index << "\t" << res_ref << "\t" << res_fit << "\n";

				vRefAligned.push_back( vRefStruct[res_ref] );
				vFitAligned.push_back( vFitStruct[res_fit] );
			}
		}
	}

	fclose( pInputFile_pairs );
}




void writeRotationMatrix( string filename ){
	FILE *pOutputFile = fopen( filename.c_str(), "w" );
	
	fprintf( pOutputFile, "*** Input structures ***\n" );
	fprintf( pOutputFile, "structure 1: " );
	fprintf( pOutputFile, "%s", s1.c_str() );
	fprintf( pOutputFile, "\n" );
	fprintf( pOutputFile, "structure 2: " );
	fprintf( pOutputFile, "%s", s2.c_str() );
	fprintf( pOutputFile, "\n" );
	fprintf( pOutputFile, "aligned structure: " );
	fprintf( pOutputFile, "%s", fit_struct.c_str() );
	fprintf( pOutputFile, "\n" );
	fprintf( pOutputFile, "\n" );
	
	fprintf( pOutputFile, "*** Rotation matrix to superpose a structure to reference structure ***\n" );
	fprintf( pOutputFile, "i         t(i)           u(i,1)        u(i,2)         u(i,3)\n" );
	
	fprintf( pOutputFile, "1   " );
	fprintf( pOutputFile, "%15.10f", t[0] );
	fprintf( pOutputFile, "%15.10f", u[0][0] );
	fprintf( pOutputFile, "%15.10f", u[0][1] );
	fprintf( pOutputFile, "%15.10f", u[0][2] );
	fprintf( pOutputFile, "\n" );
	
	fprintf( pOutputFile, "2   " );
	fprintf( pOutputFile, "%15.10f", t[1] );
	fprintf( pOutputFile, "%15.10f", u[1][0] );
	fprintf( pOutputFile, "%15.10f", u[1][1] );
	fprintf( pOutputFile, "%15.10f", u[1][2] );
	fprintf( pOutputFile, "\n" );
	
	fprintf( pOutputFile, "3   " );	
	fprintf( pOutputFile, "%15.10f", t[2] );
	fprintf( pOutputFile, "%15.10f", u[2][0] );
	fprintf( pOutputFile, "%15.10f", u[2][1] );
	fprintf( pOutputFile, "%15.10f", u[2][2] );
	fprintf( pOutputFile, "\n" );
	
	fprintf( pOutputFile, "\n" );	
	fprintf( pOutputFile, "X = t[1] + u[1][1]*x + u[1][2]*y + u[1][3]*z\n" );	
	fprintf( pOutputFile, "Y = t[2] + u[2][1]*x + u[2][2]*y + u[2][3]*z\n" );	
	fprintf( pOutputFile, "Z = t[3] + u[3][1]*x + u[3][2]*y + u[3][3]*z\n" );	
	
	fclose( pOutputFile );
}




void setBsetRotationMatrix(){
	for( int i = 0 ; i < 3 ; i++ ){
		 t_best[i] = t[i];
		 
		 u_best[i][0] = u[i][0];
		 u_best[i][1] = u[i][1];
		 u_best[i][2] = u[i][2];
	}
}




void CopyBestRotationMatrix(){
	for( int i = 0 ; i < 3 ; i++ ){
		 t[i] = t_best[i];
		 
		 u[i][0] = u_best[i][0];
		 u[i][1] = u_best[i][1];
		 u[i][2] = u_best[i][2];
	}
}




// ===============================================================
// This is a function to accurately identify aligned residue pairs
// by solving Linear Sum Assignment Problem (LSAP)
// Reference:
// Derigs,U. (1985) The shortest augumenting path method for solving assignment problems - Motivation and computational experience. In: Monma,C.L. (ed.)
// Algorithms and Software for Optimization. Baltzer, Basel, pp. 57–102.
// 
// Modified by H. S. Lee
// ==============================================================

void assignAlignedResiduePairs( vector< Residue * > vStruct1, vector< Residue * > vStruct2, int align_mode  ){
	int n_1 = vStruct1.size();	 // vRefStruct.size() <= vFitStruct.size()
	int n_2 = vStruct2.size();
	int n;
	
	if( n_1 >= n_2 )
		n = n_1;
	else
		n = n_2;
		
	float d_2 = 0;
	float d0_2_Ring = d0_Ring * d0_Ring;
	
	for( int i = 1 ; i <= n_1 ; i++ ){
		for( int j = 1 ; j <= n_2 ; j++ ){
			cost[i][j] = 2;
		}
	}
	
	if( align_mode == 1 ){
		for( int i = 0 ; i < n_1 ; i++ ){
			for( int j = 0 ; j < n_2 ; j++ ){
				d_2 = pow( vStruct1[i]->R[0] - vStruct2[j]->R_sup[0], 2 ) + pow( vStruct1[i]->R[1] - vStruct2[j]->R_sup[1], 2 ) + pow( vStruct1[i]->R[2] - vStruct2[j]->R_sup[2], 2 );
            	score[i+1][j+1] = 1 / ( 1 + d_2 / d0_2_Ring );            	
        	}
		}
	}	
	else if( align_mode == 2 ){
		for( int i = 0 ; i < n_1 ; i++ ){
			for( int j = 0 ; j < n_2 ; j++ ){
				d_2 = pow( vStruct1[i]->R_sup[0] - vStruct2[j]->R[0], 2 ) + pow( vStruct1[i]->R_sup[1] - vStruct2[j]->R[1], 2 ) + pow( vStruct1[i]->R_sup[2] - vStruct2[j]->R[2], 2 );
            	score[i+1][j+1] = 1 / ( 1 + d_2 / d0_2_Ring );            	
        	}
		}
	}
	
		
	for( int i = 1 ; i <= n_1 ; i++ ){
		for( int j = 1 ; j <= n_2 ; j++ ){
            cost[i][j] = 2 - score[i][j];
        }
	}
	
	
	double eps = 1E-10;
	double sup = 1E+9;
	double cc, d, ui, vgl, vj, z;
	double dminus[2000];
	double dplus[2000];
	double ys[2000];
	double yt[2000];
	int i, j, u, w, ind, index, j0;
	int zeile[2000];
	int vor[2000];
	int label[2000];
	
	
	for( i = 1 ; i <= n ; i++ ){
		zeile[i] = 0;
		spalte[i] = 0;
		vor[i] = 0;
		ys[i] = 0.0E+00;
		yt[i] = 0.0E+00;
	}
	
	for( i = 1 ; i <= n_2 ; i++ ){
		for( j = 1 ; j <= n_1 ; j++ ){
			cc = cost[j][i];
			
			if( j != 1 ){
				if( ( cc - ui ) >= eps ){
				}
				else{
					ui = cc;
					j0 = j;
				}
			}
			else{
				ui = cc;
				j0 = j;
			}
		}	
			
		ys[i] = ui;
					
		if( zeile[j0] == 0 ){
			zeile[j0] = i;
			spalte[i] = j0;
		}
	}
	
	for( j = 1 ; j <= n_1 ; j++ ){
		yt[j] = 0;
		
		if( zeile[j] == 0 ){
			yt[j] = sup; 
		}
	}
	
	
	for( i = 1 ; i <= n_2 ; i++ ){
		ui = ys[i];
		
		for( j = 1 ; j <= n_1 ; j++ ){
			vj = yt[j];
			
			if( eps < vj ){
				cc = cost[j][i] - ui;
				
				if( cc + eps < vj ){				
					yt[j] = cc;
					vor[j] = i;
				}
			}
		}
	}
	
	
	for( j = 1 ; j <= n_1 ; j++ ){
		i = vor[j];
		
		if( i != 0 ){
			if( spalte[i] == 0 ){
				spalte[i] = j;
                zeile[j] = i;
			}
		}		
	}
	
	
	for( i = 1 ; i <= n_2 ; i++ ){
		if( spalte[i] == 0 ){
			ui = ys[i];
			
			for( j = 1 ; j <= n_1 ; j++ ){
				if ( zeile[j] == 0 ){
					cc = cost[j][i];
              		
              		if ( cc - ui - yt[j] <= -eps ){
              			spalte[i] = j;
               			zeile[j] = i;
              		} 
				}
			}		
		}
	}
	
	
	for( u = 1 ; u <= n_2 ; u++ ){
		if( spalte[u] == 0 ){
			for( i = 1 ; i <= n_1 ; i++ ){
				vor[i] = u;
          		label[i] = 0;
          		dplus[i] = sup;
          		dminus[i] = cost[i][u] - ys[u] - yt[i];
			}
			
			dplus[u] = 0.0E+00;
			
			while( 1 ){
				d = sup;
				index = 0;
			
				for( i = 1 ; i <= n_1 ; i++ ){
					if( label[i] == 0 ){
						if( dminus[i] + eps < d ){
							d = dminus[i];
              				index = i;
						}
					}
				}
			
				if( index == 0 ){
					//cout << "LSAPR - Fatal error! No unlabeled node with DMINUS" << "\n";
            		return;
				}
			
				if( zeile[index] <= 0 ){
					while( 1 ){
						w = vor[index];
						zeile[index] = w;
        				ind = spalte[w];
        				spalte[w] = index;
        			
        				if( w != u ){
          					index = ind;
        				}
        				else{
        					break;
        				}
					}
				
					for( i = 1 ; i <= n_2 ; i++ ){
          				if( dplus[i] != sup ){
           					ys[i] = ys[i] + d - dplus[i];
          				}

          				if( dminus[i] + eps < d ){
            				yt[i] = yt[i] + dminus[i] - d;
          				}
        			}
        			
      				break;
				}
				else{
					label[index] = 1;
        			w = zeile[index];
        			dplus[w] = d;

        			for( i = 1 ; i <= n_1 ; i++ ){
          				if( label[i] == 0 ){
            				vgl = d + cost[i][w] - ys[w] - yt[i];
            			
            				if( vgl + eps < dminus[i] ){
              					dminus[i] = vgl;
              					vor[i] = w;
              				}
            			}
          			}
				}							
			}
		}
	}	
}	




void executeAssignAlignedResiduePairs( vector< Residue * > vRefStruct, vector< Residue * > vFitStruct, int align_mode ){
	if( align_mode == 1 )
		assignAlignedResiduePairs( vRefStruct, vFitStruct, align_mode );	
	else if( align_mode == 2 )
		assignAlignedResiduePairs( vFitStruct, vRefStruct, align_mode );	
}		




void setAlignedResiduePairs( int align_mode, int n_ref, int n_fit ){
	for( int i = 0 ; i < 20 ; i++ ){
		alignedRes[i] = -1;
	}
	
	if( align_mode == 1 ){		
		for( int i = 1 ; i <= n_fit ; i++ ){
			if( spalte[i] != 0 ){
				alignedRes[spalte[i]-1] = i-1;
			}
		}
	}	
	else if( align_mode == 2 ){		
		for( int i = 1 ; i <= n_ref ; i++ ){
			if( spalte[i] != 0 ){
				alignedRes[i-1] = spalte[i]-1;
			}
		}	
	}
}




void filterAlignedResiduePairs( vector< Residue * > vRefStruct, vector< Residue * > vFitStruct, float cutoff ){
	float dist;
	int n_ref = vRefStruct.size();
	
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			dist = sqrt( pow( vRefStruct[i]->R[0] - vFitStruct[alignedRes[i]]->R_sup[0], 2 ) + pow( vRefStruct[i]->R[1] - vFitStruct[alignedRes[i]]->R_sup[1], 2 ) + pow( vRefStruct[i]->R[2] - vFitStruct[alignedRes[i]]->R_sup[2], 2 ) );
			
		    if( dist > cutoff ){
		    	alignedRes[i] = -1;
		    }
		}
	}
}




float getCoverage( int n_ref ){
	int num_ali = 0;
	
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 )
			num_ali++;			
	}
	
	return num_ali/(float)n_ref;
}




void setAlignedResiduePairsWithBestScore( int n_ref ){
	for( int i = 0 ; i < 20 ; i++ ){
		alignedpair[i] = -1;
	}
	
	for( int i = 0 ; i < n_ref ; i++ ){
		alignedpair[i] = alignedRes[i];
	}
}




float getGSScore( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct, int N_target_mode ){	
	int n_ref = vRefStruct.size();	 
	int n_fit = vFitStruct.size();
	int last_resSeq_ref = vRefStruct[n_ref-1]->resSeq;
	int last_resSeq_fit = vFitStruct[n_fit-1]->resSeq;
	int n_min, n_max;
	
	if( last_resSeq_ref < last_resSeq_fit ){
		n_min = last_resSeq_ref;
		n_max = last_resSeq_fit;
	}
	else{
		n_min = last_resSeq_fit;
		n_max = last_resSeq_ref;
	}
	
	int N_target;
	
	if( N_target_mode == 1 ){
		N_target = n_max * 2 - 1;
	}
	else if( N_target_mode == 2 ){
		N_target = n_min * 2 - 1;
	}
		
	float d_Ring = 0;
	float d_OG = 0;
 	float gs_score = 0; 
 	
 	float d0_2_OG = d0_OG * d0_OG;
 	float d0_2_Ring = d0_Ring * d0_Ring;	
		
		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			d_Ring = 0;
			
			if( ( vRefStruct[i]->hasOG == 1 ) && ( vFitStruct[alignedRes[i]]->hasOG == 1 ) ){
				for( int k = 0 ; k < 6 ; k++ )
					d_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
            	d_Ring = sqrt( d_Ring / 6 );	
            	
            	d_OG = sqrt( pow( vRefStruct[i]->OG[0] - vFitStruct[alignedRes[i]]->OG_sup[0], 2 ) + pow( vRefStruct[i]->OG[1] - vFitStruct[alignedRes[i]]->OG_sup[1], 2 ) + pow( vRefStruct[i]->OG[2] - vFitStruct[alignedRes[i]]->OG_sup[2], 2 ) );
	
	        	gs_score += 1 / ( 1 + ( d_Ring * d_Ring / d0_2_Ring ) ) + 1 / ( 1 + ( d_OG * d_OG / d0_2_OG ) );        
			}
			else{
				for( int k = 0 ; k < 6 ; k++ )
					d_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
            	d_Ring = sqrt( d_Ring / 6 );	
            	
            	gs_score += 1 / ( 1 + ( d_Ring * d_Ring / d0_2_Ring ) );            	
			}            
		}
	}
		
	gs_score = ( 1 / (float)N_target ) * gs_score;
		

	return gs_score;
}




float getNormalizedGSScore( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct, int N_target_mode ){	
	int n_ref = vRefStruct.size();	 
	int n_fit = vFitStruct.size();
	int last_resSeq_ref = vRefStruct[n_ref-1]->resSeq;
	int last_resSeq_fit = vFitStruct[n_fit-1]->resSeq;
	int n_min, n_max;
	
	if( last_resSeq_ref < last_resSeq_fit ){
		n_min = last_resSeq_ref;
		n_max = last_resSeq_fit;
	}
	else{
		n_min = last_resSeq_fit;
		n_max = last_resSeq_ref;
	}
	
	int N_target;
	d0_OG = 1.36 * pow( n_max - 2.0, 1.0/2.0 ) - 0.75;      
	d0_Ring = 1.64 * pow( n_max - 2.0, 1.0/2.0 ) - 0.30;
	
	if( N_target_mode == 1 ){
		N_target = n_max * 2 - 1;
	}
	else if( N_target_mode == 2 ){
		N_target = n_min * 2 - 1;
	}
	else if( N_target_mode == 3 ){
		N_target = last_resSeq_ref * 2 - 1;
	}
			
	float d_Ring = 0;
	float d_OG = 0;
 	float gs_score = 0;  	
	
	float d0_2_OG = d0_OG * d0_OG;
 	float d0_2_Ring = d0_Ring * d0_Ring;	
		
		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			d_Ring = 0;
			
			if( ( vRefStruct[i]->hasOG == 1 ) && ( vFitStruct[alignedRes[i]]->hasOG == 1 ) ){
				for( int k = 0 ; k < 6 ; k++ )
					d_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
            	d_Ring = sqrt( d_Ring / 6 );	
            	
            	d_OG = sqrt( pow( vRefStruct[i]->OG[0] - vFitStruct[alignedRes[i]]->OG_sup[0], 2 ) + pow( vRefStruct[i]->OG[1] - vFitStruct[alignedRes[i]]->OG_sup[1], 2 ) + pow( vRefStruct[i]->OG[2] - vFitStruct[alignedRes[i]]->OG_sup[2], 2 ) );
	
	        	gs_score += 1 / ( 1 + ( d_Ring * d_Ring / d0_2_Ring ) ) + 1 / ( 1 + ( d_OG * d_OG / d0_2_OG ) );        
			}
			else{
				for( int k = 0 ; k < 6 ; k++ )
					d_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
            	d_Ring = sqrt( d_Ring / 6 );	
            	
            	gs_score += 1 / ( 1 + ( d_Ring * d_Ring / d0_2_Ring ) );            	
			}            
		}
	}
		
	gs_score = ( 1 / (float)N_target ) * gs_score;
		

	return gs_score;
}




float getAverageDforOG( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct ){	
	int n_ref = vRefStruct.size();	
	int n_fit = vFitStruct.size();
	float avg_d_OG = 0;
    int N_align = 0;
    		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			if( ( vRefStruct[i]->hasOG == 1 ) && ( vFitStruct[alignedRes[i]]->hasOG == 1 ) ){    
				N_align++;        	
            	avg_d_OG += sqrt( pow( vRefStruct[i]->OG[0] - vFitStruct[alignedRes[i]]->OG_sup[0], 2 ) + pow( vRefStruct[i]->OG[1] - vFitStruct[alignedRes[i]]->OG_sup[1], 2 ) + pow( vRefStruct[i]->OG[2] - vFitStruct[alignedRes[i]]->OG_sup[2], 2 ) );
			}			      
		}
	}

	return avg_d_OG / (float)N_align;
}




float getAverageDforRing( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct ){	
	int n_ref = vRefStruct.size();	
	int n_fit = vFitStruct.size();
	float d_Ring = 0;
	float avg_d_Ring = 0;
    int N_align = 0;
    		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			N_align++;   
			     	
			for( int k = 0 ; k < 6 ; k++ )
				d_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
           	d_Ring = sqrt( d_Ring / 6 );	
           	
           	avg_d_Ring += d_Ring;
		}
	}

	return avg_d_Ring / (float)N_align;
}




int getNumAlignedResidues( int n_ref ){
	int N_align = 0;
    		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			N_align++;   
		}
	}
	
	return N_align;
}




float getRMSD( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct ){	
	int n_ref = vRefStruct.size();	 
	int n_fit = vFitStruct.size();
	int N_target;
	
	float d2_Ring = 0;
	float d2_OG = 0;
	int N_align = 0;
 	float rmsd = 0; 
 		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			d2_Ring = 0;
			
			if( ( vRefStruct[i]->hasOG == 1 ) && ( vFitStruct[alignedRes[i]]->hasOG == 1 ) ){
				for( int k = 0 ; k < 6 ; k++ )
					d2_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
                d2_Ring = d2_Ring / 6;
                
	           	d2_OG = pow( vRefStruct[i]->OG[0] - vFitStruct[alignedRes[i]]->OG_sup[0], 2 ) + pow( vRefStruct[i]->OG[1] - vFitStruct[alignedRes[i]]->OG_sup[1], 2 ) + pow( vRefStruct[i]->OG[2] - vFitStruct[alignedRes[i]]->OG_sup[2], 2 );
	
	        	rmsd += d2_Ring + d2_OG;     
	        	N_align = N_align + 2;   
			}
			else{
				for( int k = 0 ; k < 6 ; k++ )
					d2_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedRes[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedRes[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedRes[i]]->Ring_sup[k][2], 2 );
            
                d2_Ring = d2_Ring / 6;
                
            	rmsd += d2_Ring;        
            	N_align++;    	
			}            
		}
	}
	
	rmsd = sqrt( rmsd / (float)N_align );
		

	return rmsd;
}




float getRMSDForBestGSScore( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct ){	
	int n_ref = vRefStruct.size();	 
	int n_fit = vFitStruct.size();
	int N_target;
	
	float d2_Ring = 0;
	float d2_OG = 0;
	int N_align = 0;
 	float rmsd = 0; 
 		
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedpair[i] != -1 ){
			d2_Ring = 0;
			
			if( ( vRefStruct[i]->hasOG == 1 ) && ( vFitStruct[alignedpair[i]]->hasOG == 1 ) ){
				for( int k = 0 ; k < 6 ; k++ )
					d2_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedpair[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedpair[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedpair[i]]->Ring_sup[k][2], 2 );
            
                d2_Ring = d2_Ring / 6;
                
	           	d2_OG = pow( vRefStruct[i]->OG[0] - vFitStruct[alignedpair[i]]->OG_sup[0], 2 ) + pow( vRefStruct[i]->OG[1] - vFitStruct[alignedpair[i]]->OG_sup[1], 2 ) + pow( vRefStruct[i]->OG[2] - vFitStruct[alignedpair[i]]->OG_sup[2], 2 );
	
	        	rmsd += d2_Ring + d2_OG;  
	        	N_align = N_align + 2;       
			}
			else{
				for( int k = 0 ; k < 6 ; k++ )
					d2_Ring += pow( vRefStruct[i]->Ring[k][0] - vFitStruct[alignedpair[i]]->Ring_sup[k][0], 2 ) + pow( vRefStruct[i]->Ring[k][1] - vFitStruct[alignedpair[i]]->Ring_sup[k][1], 2 ) + pow( vRefStruct[i]->Ring[k][2] - vFitStruct[alignedpair[i]]->Ring_sup[k][2], 2 );
            
                d2_Ring = d2_Ring / 6;
                
            	rmsd += d2_Ring;   
            	N_align++;          	
			}            
		}
	}
	
	rmsd = sqrt( rmsd / (float)N_align );
		

	return rmsd;
}




void setFragmentLinear( vector< Residue * > &vStruct, vector< Fragment * > &vFragment ){
	int i_index, j_index, k_index;
	Fragment *frag;
	
	int n = vStruct.size();
	
	for( int k = n - 1 ; k >= 2 ; k-- ){
		i_index = j_index = k_index = -1;
		
		k_index = k;
		
		if( vStruct[k_index]->resSeq_OG_linked != 0 ){
			j_index = vStruct[k_index]->resSeq_OG_linked - 1;
				
			if( vStruct[j_index]->resSeq_OG_linked != 0 ){
				i_index = vStruct[j_index]->resSeq_OG_linked - 1;
					
				frag = new Fragment();
				frag->index[0] = i_index;
				frag->index[1] = j_index;	
				frag->index[2] = k_index;	
					
				vFragment.push_back( frag );	
			}
		}
	}					
}




void setLinkedResidueInf( vector< Residue * > &vStruct ){ 
	for( int i = 0 ; i < 20 ; i++ ){
		num_linkages[i] = 0;
		linked_residues[i][0] = -1;
		linked_residues[i][1] = -1;
	}
	
	int n = vStruct.size();
	
	for( int i = 0 ; i < n ; i++ ){
		if( vStruct[i]->resSeq_OG_linked != 0 ){
			if( num_linkages[vStruct[i]->resSeq_OG_linked-1] < 2 ){
				num_linkages[vStruct[i]->resSeq_OG_linked-1]++;
				
				if( linked_residues[vStruct[i]->resSeq_OG_linked-1][0] == -1 )
					linked_residues[vStruct[i]->resSeq_OG_linked-1][0] = i; 
				else
					linked_residues[vStruct[i]->resSeq_OG_linked-1][1] = i; 
			}			
		}
	} 
}




void setFragmentBranched( vector< Residue * > &vStruct, vector< Fragment * > &vFragment ){
	int i_index, j_index, k_index, l_index;
	Fragment *frag;
	int n = vStruct.size();
	
	for( int i = 0 ; i < n ; i++ ){
		if( ( num_linkages[i] == 2 ) && ( vStruct[i]->resSeq_OG_linked != 0 ) ){
			i_index = vStruct[i]->resSeq_OG_linked - 1;
			j_index = i;
			k_index = linked_residues[i][0];
			l_index = linked_residues[i][1];
			
			//cout << i_index << " " << j_index << " " << k_index << " " << l_index << "\n";
			
			frag = new Fragment();
			frag->index[0] = i_index;
			frag->index[1] = j_index;	
			frag->index[2] = k_index;	
			frag->index[3] = l_index;	
					
			vFragment.push_back( frag );	
			
					
			frag = new Fragment();
			frag->index[0] = i_index;
			frag->index[1] = j_index;	
			frag->index[2] = l_index;	
			frag->index[3] = k_index;	
					
			vFragment.push_back( frag );	
		}		
	}
}




void setAlignedAtomPairsUsingLinearFragments( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct, Fragment * ref_frag, Fragment * fit_frag, vector< Atom * > &vRefAligned, vector< Atom * > &vFitAligned ){
	Atom * atom;
		
	for( int i = 0 ; i < 3 ; i++ ){
		atom = new Atom();
		atom->R[0] = vRefStruct[ref_frag->index[i]]->R[0];
		atom->R[1] = vRefStruct[ref_frag->index[i]]->R[1];
		atom->R[2] = vRefStruct[ref_frag->index[i]]->R[2];		
		vRefAligned.push_back( atom );
		
		atom = new Atom();
		atom->R[0] = vFitStruct[fit_frag->index[i]]->R[0];
		atom->R[1] = vFitStruct[fit_frag->index[i]]->R[1];
		atom->R[2] = vFitStruct[fit_frag->index[i]]->R[2];		
		vFitAligned.push_back( atom );		
	}			
}




void setAlignedAtomPairsUsingBranchedFragments( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct, Fragment * ref_frag, Fragment * fit_frag, vector< Atom * > &vRefAligned, vector< Atom * > &vFitAligned ){
	Atom * atom;
		
	for( int i = 0 ; i < 4 ; i++ ){
		atom = new Atom();
		atom->R[0] = vRefStruct[ref_frag->index[i]]->R[0];
		atom->R[1] = vRefStruct[ref_frag->index[i]]->R[1];
		atom->R[2] = vRefStruct[ref_frag->index[i]]->R[2];		
		vRefAligned.push_back( atom );
		
		atom = new Atom();
		atom->R[0] = vFitStruct[fit_frag->index[i]]->R[0];
		atom->R[1] = vFitStruct[fit_frag->index[i]]->R[1];
		atom->R[2] = vFitStruct[fit_frag->index[i]]->R[2];		
		vFitAligned.push_back( atom );		
	}			
}




void setAlignedAtomPairsUsingAlignedResArray( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct, vector< Atom * > &vRefAligned, vector< Atom * > &vFitAligned ){
	Atom * atom;
	int n_ref = vRefStruct.size();
	
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedRes[i] != -1 ){
			atom = new Atom();				
			atom->R[0] = vRefStruct[i]->R[0];
			atom->R[1] = vRefStruct[i]->R[1];
			atom->R[2] = vRefStruct[i]->R[2];				
			vRefAligned.push_back( atom );
				
			atom = new Atom();				
			atom->R[0] = vFitStruct[alignedRes[i]]->R[0];
			atom->R[1] = vFitStruct[alignedRes[i]]->R[1];
			atom->R[2] = vFitStruct[alignedRes[i]]->R[2];				
			vFitAligned.push_back( atom );
		}
	}	
}




void showAlignedResiduePairs( vector< Residue * > &vRefStruct, vector< Residue * > &vFitStruct  ){
	int n_ref = vRefStruct.size();
	float d_Ring = 0;
	
	cout << "\n";
	cout << "*** Seq. Alignment (s1 --- s2) ***" << "\n";
	
	for( int i = 0 ; i < n_ref ; i++ ){
		if( alignedpair[i] != -1 ){
			cout << vRefStruct[i]->resName << " ";
			
			if( vRefStruct[i]->resSeq < 10 )
				cout << " " << vRefStruct[i]->resSeq;
			else
				cout << vRefStruct[i]->resSeq;
				
			cout << " --- ";
			cout << vFitStruct[alignedpair[i]]->resName << " ";
			
			if( vFitStruct[alignedpair[i]]->resSeq < 10 )
				cout << " " << vFitStruct[alignedpair[i]]->resSeq << "\n";
			else
				cout << vFitStruct[alignedpair[i]]->resSeq << "\n";
		}
		else{
			cout << vRefStruct[i]->resName << " ";
			
			if( vRefStruct[i]->resSeq < 10 )
				cout << " " << vRefStruct[i]->resSeq << "\n";
			else
				cout << vRefStruct[i]->resSeq << "\n";
		}
	}	
}




void clearEgdeVector( vector< Edge * > &vEdge ){
	for( int i = 0 ; i < vEdge.size() ; i++ )	    
		delete vEdge[i];
				
	vEdge.clear();			
	vEdge.resize( 0 ); 
}




void clearPairsVector( vector< int > &vPairs ){
	vPairs.clear();			
	vPairs.resize( 0 ); 	
}




void clearPairVector( vector< Pair * > &vPair ){
	for( int i = 0 ; i < vPair.size() ; i++ )	    
		delete vPair[i];

	vPair.clear();			
	vPair.resize( 0 ); 
}




void clearAlignedResidueVector( vector< Atom * > &vAligned ){
	vAligned.clear();			
	vAligned.resize( 0 ); 
}




void clearAlignedResidueVectorFragment( vector< Atom * > &vAligned ){
	for( int i = 0 ; i < vAligned.size() ; i++ )	    
		delete vAligned[i];
		
	vAligned.clear();			
	vAligned.resize( 0 ); 
}




void clearAtomVector( vector< Atom * > &vStruct ){
	for( int i = 0 ; i < vStruct.size() ; i++ )	    
		delete vStruct[i];

	vStruct.clear();			
	vStruct.resize( 0 ); 	
}




void clearResidueVector( vector< Residue * > &vStruct ){
	for( int i = 0 ; i < vStruct.size() ; i++ ){	    
		delete vStruct[i];
	}

	vStruct.clear();			
	vStruct.resize( 0 ); 	
}




void clearFragmentVector( vector< Fragment * > &vFragment ){
	for( int i = 0 ; i < vFragment.size() ; i++ ){
		delete vFragment[i];
	}

	vFragment.clear();			
	vFragment.resize( 0 ); 
}










// ***********************************
// ********** GS-align MAIN **********
// ***********************************

int main( int argc,char *argv[] ){
	parseCommandLine( argc, argv ); 
	
	
	
	// *** Variables ***
	float rmsd = 0;
	float best_rmsd = 100;
	float rmsd_ali = 0;
	float best_rmsd_ali = 100;
	float coverage_ali = 0;
	float best_coverage_ali = 0;
	float gs_score = 0;
	float best_gs_score = 0;
	float best_dist_tolerance = 0;	
	float avg_d_OG = 0;
	float avg_d_Ring = 0;
	
	float d_OG, d_Ring;
	float best_d_OG, best_d_Ring;
	
	int n_ref, n_fit;
	int n_aligned;
    int align_mode = 0;	
	
	vector< Atom * > vRefStruct;    
	vector< Atom * > vFitStruct;
	
	vector< Pair * > vPair;
	Pair *p; 
	
	vector< Edge * > vEdge;
	Edge *e; 
	
	vector< int > vPairs;
	
	vector< Atom * > vRefAligned;
	vector< Atom * > vFitAligned;
	
	vector< Residue * > vRefStruct_s; 
	vector< Residue * > vFitStruct_s;
	
	vector< Fragment * > vRefFragment;
	vector< Fragment * > vFitFragment;
	
	
	
	
	int num_res_s1 = getNumberOfResidues( s1 );
	int num_res_s2 = getNumberOfResidues( s2 );
	
	if( ( num_res_s1 < 2 ) || ( num_res_s2 < 2 ) ){
		cout << "The number of monosaccharides should be >= 2." << "\n";
		return 0;
	}
		 
		 
			
	
	
	// =============================
	// *** read input structures ***
	// =============================
	
	ref_struct = s1;
	fit_struct = s2;
	
	readOGAndRingCentroidFromPDBFile( ref_struct, vRefStruct );
	readOGAndRingCentroidFromPDBFile( fit_struct, vFitStruct );
		
	
	readStructureForScoring( ref_struct, vRefStruct_s  );
	setGlycosidicOxygens( ref_struct, vRefStruct_s );
	
	readStructureForScoring( fit_struct, vFitStruct_s  );
	setGlycosidicOxygens( fit_struct, vFitStruct_s );

    n_ref = vRefStruct_s.size();
	n_fit = vFitStruct_s.size();
	
	if( n_ref >= n_fit )
		align_mode = 1;
	else
		align_mode = 2;


    // *** generate all possible vertices pairs for maximum clique search ***
	GeneratePairs( vRefStruct, vFitStruct, vPair );
	writePairs( vPair );	
	
	
	
		
	// ================================================
	// *** Maximum clique-based structure alignment ***
	// ================================================
	
	for( int i = 0 ; i < 5 ; i++ ){
	    // *** generate product graph ***
	    dist_tolerance = dist_tolerance + i * 0.5;
		GenerateEdges( vRefStruct, vFitStruct, vPair, vEdge, dist_tolerance );
		writeProductGraph( vPair.size(), vEdge );



	    // *** identify maximum clique ***
		int max_clique_result = identifyMaxClique( vPairs );
	
		if( max_clique_result == 0 ){
			gs_score = 0;
		}
		else{		
			setMaxCliqueAlignedNodes( vRefStruct, vFitStruct, vRefAligned, vFitAligned, vPairs );
			setRotationMatrix( vRefAligned, vFitAligned );	
			setSuperposedCoordinatesForFitStructure( vFitStruct_s );

			executeAssignAlignedResiduePairs( vRefStruct_s, vFitStruct_s, align_mode );	
			setAlignedResiduePairs( align_mode, n_ref, n_fit );
			
			gs_score = getNormalizedGSScore( vRefStruct_s, vFitStruct_s, N_target_option );	
			rmsd = getRMSD( vRefStruct_s, vFitStruct_s );			
			
			filterAlignedResiduePairs( vRefStruct_s, vFitStruct_s, 5.0 );
			coverage_ali = getCoverage( vRefStruct_s.size() );	
			rmsd_ali = getRMSD( vRefStruct_s, vFitStruct_s );	
		}
		
		if( gs_score > best_gs_score ){
			best_dist_tolerance = dist_tolerance;
			
			best_gs_score = gs_score;
			best_rmsd = rmsd;			
			best_coverage_ali = coverage_ali;
			best_rmsd_ali = rmsd_ali;
			
			setBsetRotationMatrix();	
			setAlignedResiduePairsWithBestScore( n_ref );
		}
		
		clearEgdeVector( vEdge );
		clearPairsVector( vPairs );		
		clearAlignedResidueVector( vRefAligned );
		clearAlignedResidueVector( vFitAligned );			
	}
	
	
	
	
	
	// =======================================================
	// ***************** fragment alignments *****************
	// =======================================================
	
	// Alignment Using Linear Fragments
	// I--J--K
	// the length of glycan fragment: 3 
	
	setFragmentLinear( vRefStruct_s, vRefFragment );
	setFragmentLinear( vFitStruct_s, vFitFragment );	
	
	for( int i = 0 ; i < vRefFragment.size() ; i++ ){
		for( int j = 0 ; j < vFitFragment.size() ; j++ ){
			setAlignedAtomPairsUsingLinearFragments( vRefStruct_s, vFitStruct_s, vRefFragment[i], vFitFragment[j], vRefAligned, vFitAligned );
			setRotationMatrix( vRefAligned, vFitAligned );				
			setSuperposedCoordinatesForFitStructure( vFitStruct_s );
			
			executeAssignAlignedResiduePairs( vRefStruct_s, vFitStruct_s, align_mode );	
			setAlignedResiduePairs( align_mode, n_ref, n_fit );
			filterAlignedResiduePairs( vRefStruct_s, vFitStruct_s, 8.0 );	
			clearAlignedResidueVectorFragment( vRefAligned );
			clearAlignedResidueVectorFragment( vFitAligned );
            
			setAlignedAtomPairsUsingAlignedResArray( vRefStruct_s, vFitStruct_s, vRefAligned, vFitAligned );
			setRotationMatrix( vRefAligned, vFitAligned );	
			setSuperposedCoordinatesForFitStructure( vFitStruct_s );
			executeAssignAlignedResiduePairs( vRefStruct_s, vFitStruct_s, align_mode );	
			setAlignedResiduePairs( align_mode, n_ref, n_fit );
			
			gs_score = getNormalizedGSScore( vRefStruct_s, vFitStruct_s, N_target_option );
			rmsd = getRMSD( vRefStruct_s, vFitStruct_s );
			
			filterAlignedResiduePairs( vRefStruct_s, vFitStruct_s, 5.0 );	
			coverage_ali = getCoverage( vRefStruct_s.size() );	
			rmsd_ali = getRMSD( vRefStruct_s, vFitStruct_s );		
			
			clearAlignedResidueVectorFragment( vRefAligned );
			clearAlignedResidueVectorFragment( vFitAligned );		
			
			if( gs_score > best_gs_score ){
				best_dist_tolerance = dist_tolerance;
			
				best_gs_score = gs_score;
				best_rmsd = rmsd;			
				best_coverage_ali = coverage_ali;
				best_rmsd_ali = rmsd_ali;
			
				setBsetRotationMatrix();	
				setAlignedResiduePairsWithBestScore( n_ref );
			}
		}
	}
	
	clearFragmentVector( vRefFragment );
	clearFragmentVector( vFitFragment );	
	
	
	// Alignment Using Tripodal Fragments
	// K     L
	//  \ J /
	//    |
	//    I
	// the length of glycan fragment: 4 
	
	setLinkedResidueInf( vRefStruct_s );
	setFragmentBranched( vRefStruct_s, vRefFragment );
	setLinkedResidueInf( vFitStruct_s );
	setFragmentBranched( vFitStruct_s, vFitFragment );
		
	for( int i = 0 ; i < vRefFragment.size() ; i++ ){
		for( int j = 0 ; j < vFitFragment.size() ; j++ ){
			setAlignedAtomPairsUsingBranchedFragments( vRefStruct_s, vFitStruct_s, vRefFragment[i], vFitFragment[j], vRefAligned, vFitAligned );
			setRotationMatrix( vRefAligned, vFitAligned );	
			setSuperposedCoordinatesForFitStructure( vFitStruct_s );
			
			executeAssignAlignedResiduePairs( vRefStruct_s, vFitStruct_s, align_mode );	
			setAlignedResiduePairs( align_mode, n_ref, n_fit );
			filterAlignedResiduePairs( vRefStruct_s, vFitStruct_s, 8.0 );
			clearAlignedResidueVectorFragment( vRefAligned );
			clearAlignedResidueVectorFragment( vFitAligned );
            
			setAlignedAtomPairsUsingAlignedResArray( vRefStruct_s, vFitStruct_s, vRefAligned, vFitAligned );
			setRotationMatrix( vRefAligned, vFitAligned );	
			setSuperposedCoordinatesForFitStructure( vFitStruct_s );
			executeAssignAlignedResiduePairs( vRefStruct_s, vFitStruct_s, align_mode );	
			setAlignedResiduePairs( align_mode, n_ref, n_fit );
			
			gs_score = getNormalizedGSScore( vRefStruct_s, vFitStruct_s, N_target_option );
			rmsd = getRMSD( vRefStruct_s, vFitStruct_s );
			
			filterAlignedResiduePairs( vRefStruct_s, vFitStruct_s, 5.0 );	
			coverage_ali = getCoverage( vRefStruct_s.size() );	
			rmsd_ali = getRMSD( vRefStruct_s, vFitStruct_s );		
			
			clearAlignedResidueVectorFragment( vRefAligned );
			clearAlignedResidueVectorFragment( vFitAligned );		
			
			if( gs_score > best_gs_score ){
				best_dist_tolerance = dist_tolerance;
			
				best_gs_score = gs_score;
				best_rmsd = rmsd;			
				best_coverage_ali = coverage_ali;
				best_rmsd_ali = rmsd_ali;
			
				setBsetRotationMatrix();	
				setAlignedResiduePairsWithBestScore( n_ref );
			}
		}
	}
	
	clearFragmentVector( vRefFragment );
	clearFragmentVector( vFitFragment );	
	
	
	
	
	
	
	
	// ====================
	// *** print scores ***
	// ====================
	
	if( output_option == 1 ){
		cout << "GS-score: " << best_gs_score << "\n";
		cout << "RMSD (Å): " << best_rmsd_ali << "\n";
	}
	else if( output_option == 2 ){			
		cout << "GS-score: " << best_gs_score << "\n";
		cout << "RMSD (Å): " << best_rmsd_ali << "\n";

		CopyBestRotationMatrix();
		
		writeRotationMatrix( "matrix.txt" );
		superposeFitStructure( fit_struct, "ali_struct.pdb" );	
		
		showAlignedResiduePairs( vRefStruct_s, vFitStruct_s );
	}
	
	
	
	
	
	// ====== Delete remaining pointers ======
	clearPairVector( vPair );
	clearEgdeVector( vEdge );
	clearAtomVector( vRefStruct );	
    clearAtomVector( vFitStruct );
	clearAlignedResidueVector( vRefAligned );
	clearAlignedResidueVector( vFitAligned );
	clearResidueVector( vRefStruct_s );
	clearResidueVector( vFitStruct_s );
    // =======================================
		


	return 0;
}
